''' Implementation of the Neighbour Joining Algorithm '''
from __future__ import print_function
import numpy as np
from scipy.optimize import minimize as optim

class Edge():
    def __init__(self,node1,node2,length):
        self.n1=node1
        self.n2=node2
        if (length>0):
            self.l=length
        else:
            self.l=1e-5
    def __str__(self):
        return "%s -- [ %f ] -- %s"%(self.n1.label,self.l,self.n2.label)
    
class Node():
    def __init__(self,label=''):
        self.label=label
        self.child=[] ## list of children nodes
        self.f=[]     ## distance to children
        self.parent=None ## parent node
    def newickize(self,parent_distance):
        ''' Create a newick representation of the node and its children'''
        if len( self.child) == 0:
            name = self.label
        else:
            name = ''
        if self.parent is None:
            tw=name
        else:
            tw=name+':'+str(parent_distance)
        if len(self.child)==0:
            return (tw)
        else:
            subtrees= (sub.newickize(self.f[i]) for i,sub in enumerate(self.child))
            return '(%s)%s' % (','.join(subtrees),tw)       
    def grow_from_edges(self,cur_edges):
        ''' Construct recursively a subtree from edges information '''
        ## we build a new tree: reset values
        self.child=[]
        self.f=[]
        mye=[e for e in cur_edges if e.n1 is self or e.n2 is self]
        if len(mye)==0:
            return
        for e in mye:
            try:
                assert e.l>0
            except AssertionError:
                print("Warning: Negative branch length encountered in NJ")
                raise
            if e.n1 is self:
                self.child.append(e.n2)
                self.f.append(e.l)
                e.n2.parent=self
            elif e.n2 is self:
                self.child.append(e.n1)
                self.f.append(e.l)
                e.n1.parent=self
            cur_edges.remove(e)
        for c in self.child:
            c.grow_from_edges(cur_edges)
        return
    def get_fi(self,curval,res):
        ''' Computes fi : distance from a tip to the root'''
        if len(self.child)==0:
            ## tip
            res[self.label]=curval
        else:
            for i,c in enumerate(self.child):
                c.get_fi(curval+self.f[i],res)
        return
    def distance_to_root(self,curval=0):
        res=curval
        if self.parent is not None:
            res+=self.parent.f[self.parent.child.index(self)]
            res+=self.parent.distance_to_root(curval)
            return res
        else:
            return 0
    def get_leaves(self):
        ''' Returns a list of leaves (node instances) under self '''
        if len(self.child)==0:
            res=[self]
        else:
            res=[]
            for c in self.child:
                res+=c.get_leaves()
        return res
    def mrca(self,other):
        ''' Identifies most recent common ancestor with other node'''
        try:
            assert len(self.child)==0 and len(other.child)==0
        except AssertionError:
            raise AssertionError('MRCA only applies to leaves')
        anc=self.parent
        while other not in anc.get_leaves():
            anc=anc.parent
            assert anc is not None
        return anc
        
###Optimization function to root the tree 
def qfunc(h0,fi,hi):
    return sum( (((1-fi)*h0-hi)/hi)**2)

def qfunc_2pars(pars,tree,hzy):
    h0=pars[0]
    posroot=pars[1]
    fi_tips=tree.calc_fi(posroot)
    fi_opt=np.zeros(len(fi_tips),dtype=float)
    hi_opt=np.zeros(len(fi_tips),dtype=float)
    for i,pop in enumerate(fi_tips.keys()):
        fi_opt[i]=fi_tips[pop]
        hi_opt[i]=hzy[pop]
    return(qfunc(h0,fi_opt,hi_opt))

class Rooted_Tree():
    def __init__(self):
        self.root=None
    def build_from_edges(self,edges,root_edge,outgroup=None):
        ''' Builds a tree from a list of edges
            If outgroup is specified, it allows to identify the root and is dropped from the tree.
            Else, the root is  positioned in the center of root_edge.
        '''
        current_edges=[e for e in edges if e is not root_edge]
        if outgroup:
            if root_edge.n1.label==outgroup:
                self.root=root_edge.n2
            elif root_edge.n2.label==outgroup:
                self.root=root_edge.n1
            else:
                raise ValueError('Outgroup %s not found'%outgroup)
            left_edges=[]
            for e in current_edges:
                ## find children and set their distances
                if e.n1.label==self.root.label:
                    self.root.child.append(e.n2)
                    self.root.f.append(e.l)
                    ##current_edges.remove(e)
                elif e.n2.label==self.root.label:
                    self.root.child.append(e.n1)
                    self.root.f.append(e.l)
                    ##current_edges.remove(e)
                else:
                    left_edges.append(e)
            current_edges=left_edges
        else:
            self.root=Node(label=root_edge.n1.label+'-'+root_edge.n2.label)
            self.root.child=[root_edge.n1,root_edge.n2]
            self.root.f=[0.5*root_edge.l,0.5*root_edge.l]
        for c in self.root.child:
            c.parent=self.root
            c.grow_from_edges(current_edges)
    def newick(self):
        if self.root:
            return self.root.newickize('')+';'
    def shift_root(self,posroot):
        root_edge_len=np.sum(self.root.f)
        self.root.f[0]=posroot*root_edge_len
        self.root.f[1]=(1-posroot)*root_edge_len
    def calc_fi(self,posroot=None):
        '''
        Get a dictionary of distances from tips to root.
        optional parameter posroot (0<pos<1) adjusts the position 
        of the root between its two children
        '''
        if posroot:
            self.shift_root(posroot)
        res={}
        self.root.get_fi(0,res)
        return res
    def optim_h0(self,hzy):
        fi_tips=self.calc_fi()
        fi_opt=np.zeros(len(fi_tips),dtype=float)
        hi_opt=np.zeros(len(fi_tips),dtype=float)
        for i,pop in enumerate(fi_tips.keys()):
            fi_opt[i]=fi_tips[pop]
            hi_opt[i]=hzy[pop]
        val=optim(qfunc,np.mean(hi_opt),args=(fi_opt,hi_opt))
        return val
    def optim_root(self,hzy):
        ''' Optimize root placement
        optim parameters: hzy_ancestral (0,0.5) , posroot(0,1)
        '''
        val=optim(qfunc_2pars,[0.25,0.5],args=(self,hzy),bounds=[ (0,0.5),(0,1) ] )
        return val
    def kinship(self):
        ''' Returns the Kinship matrix of the leaves '''
        leaves=self.root.get_leaves()
        n=len(leaves)
        kinship=np.zeros((n,n),dtype=float)
        for i,l1 in enumerate(leaves):
            kinship[i,i]=l1.distance_to_root()
            for j in range(i+1,n):
                anc=l1.mrca(leaves[j])
                val=anc.distance_to_root()
                if val<0:
                    print('Warning: zeroing negative values in kinship')
                    val=0
                kinship[i,j]=val
                kinship[j,i]=kinship[i,j]
        return kinship,[l.label for l in leaves]
    
class NJ():
    ''' Class Implementing the Neighbour Joining Algorithm
    '''
    def __init__(self,Distance,labels):
        ''' 
        Create an instance from a Distance Matrix n x n and a 
        list of corresponding population albels '''
        assert len(labels)==Distance.shape[0] and Distance.shape[0]==Distance.shape[1]
        self.D=Distance
        self.tips=[Node(x) for x in labels]
        self.Dnodes=[x for x in self.tips]
        self.edges=[]
        
    def _QfromD(self,D):
        ''' Calculates Q-matrix from D.
        Returns Q and (imax,jmax) the indices of smallest Q'''
        n=D.shape[0]
        rowsum=np.sum(D,axis=0)
        Q=(n-2)*D
        for i in range(n):
            Q[:,i]-=rowsum[i]
            Q[i,:]-=rowsum[i]
            Q[i,i]=0
        ## make sure the diagonal does not contain smallest value
        Q+=np.diag(n*[np.max(Q)+1])
        return Q,np.unravel_index(np.argmin(Q),Q.shape)

    def _get_new_D(self,D,pair):
        '''Calculates a new distance matrix after joining the pair (i,j) into new node u 
           and creating the 2 new corresponding edges.
           return newD
        '''
        n=D.shape[0]
        delta=np.zeros(2)
        imin,jmin=pair
        node1=self.Dnodes[imin]
        node2=self.Dnodes[jmin]
        node_intern=Node(label=node1.label+'-'+node2.label)
        delta[0]=0.5*D[imin,jmin]+(0.5/(n-2))*(np.sum(D[imin,])-np.sum(D[jmin,]))
        delta[1]=D[imin,jmin]-delta[0]
        ## create new edges
        new_edge=Edge(node1,node_intern,delta[0] >0 and delta[0] or 1e-5)
        self.edges.append(new_edge)
        new_edge=Edge(node2,node_intern,delta[1]>0 and delta[1] or 1e-5)
        self.edges.append(new_edge)

        newD=np.zeros((n-1,n-1))
        ind_left=[ i for i in range(n)]
        ind_left.remove(imin)
        ind_left.remove(jmin)
        self.Dnodes.remove(node1)
        self.Dnodes.remove(node2)
        self.Dnodes=[node_intern]+self.Dnodes
        newD[1:,1:]=D[ind_left,:][:,ind_left]
        for newi,i in enumerate(ind_left):
            newD[0,newi+1]=0.5*(D[imin,i]+D[jmin,i]-D[imin,jmin])
            newD[newi+1,0]=newD[0,newi+1]
        return newD
    
    def fit(self):
        '''
        Calculates the neighbor joining tree
        Sets up internal nodes and edges between nodes.
        '''
        curD=np.copy(self.D)
        while curD.shape[0]>2:
            # print('+++')        
            Q,pair=self._QfromD(curD)
            # print(Q,pair,sep='\n')
            newD=self._get_new_D(curD,pair)
            # print(self.edges[-1])
            # print(self.edges[-2])
            # print(newD,sep='\n')
            curD=newD
        ## terminate
        new_edge=Edge(self.Dnodes[0],self.Dnodes[1],curD[0,1])
        self.edges.append(new_edge)
        # print(self.edges[-1])

       
    
def test():
    ## Test distance matrix from wikipedia :)
    print("#### Wikipedia example ####\nsee https://en.wikipedia.org/wiki/Neighbor_joining")
    Dtest=np.array([[0,5,9,9,8],
                    [5,0,10,10,9],
                    [9,10,0,8,7],
                    [9,10,8,0,3],
                    [8,9,7,3,0]])
    labels=['a','b','c','d','e']
    my_nj=NJ(Dtest,labels)
    my_nj.fit()
    print("Edges from NJ:")
    print(*my_nj.edges,sep='\n')
    print("#### Ovis  example ####")
    Dsheep=2*np.array([[0.0, 0.052524410884, 0.479593743725, 0.15280193534],
                     [0.052524410884 , 0.0 , 0.489188722484 ,0.162238830969],
                     [0.479593743725 , 0.489188722484 , 0.0 , 0.423091249114],
                     [0.15280193534 , 0.162238830969 , 0.423091249114 ,0.0]])
    labels=['MOOA','IROA','Capra','IROO']
    my_nj=NJ(Dsheep,labels)
    my_nj.fit()
    print("Edges from NJ:")
    print(*my_nj.edges,sep='\n')
    ## remove outgroup
    outgroup='Capra'
    ## find edge from root
    root_edge=[e for e in my_nj.edges if e.n1.label==outgroup or e.n2.label==outgroup][0]
    tree=Rooted_Tree()
    tree.build_from_edges(my_nj.edges,root_edge,outgroup=outgroup)
    print("Outgroup Tree:",tree.newick())
    Dsheep_nocapra=Dsheep[(0,1,3),:][:,(0,1,3)]
    labels.remove('Capra')
    my_nj=NJ(Dsheep_nocapra,labels)
    my_nj.fit()
    print("\n")
    print(*my_nj.edges,sep='\n')
    hzy={'MOOA':0.2531,'IROA':0.2479,'IROO':0.2874}
    resid=np.inf
    root_edge_best=None
    with open('sheep.tree','w') as ftree:
        for root_e in my_nj.edges:
            the_edges=[e for e in my_nj.edges]
            tree=Rooted_Tree()
            tree.build_from_edges(the_edges,root_e)
            print(resid)
            # tree_fit=tree.optim_h0(hzy)
            tree_fit=tree.optim_root(hzy)
            print(tree_fit.x,tree_fit.fun)
            if tree_fit.x[1]>0 and tree_fit.fun<resid:
                resid=tree_fit.fun
                root_edge_best=root_e
                root_pos_best=tree_fit.x[1]
    print("Found best tree with %f accuracy"%resid)
    tree=Rooted_Tree()
    tree.build_from_edges(my_nj.edges,root_edge_best)
    tree.shift_root(root_pos_best)
    print("Optimized tree:",tree.newick())
    kin,popnames=tree.kinship()
    print(kin,*popnames)
    
if __name__=='__main__':
    test()



#!/usr/bin/env python3
import pickle
import sys
import argparse
import bz2
from hapflk import InputOutput as IO
from hapflk import popgen
import hapflk.fastphase as myfph
import numpy as np
from multiprocessing import cpu_count


class SNP(object):
    """ Pickable simplified SNP class"""
    def __init__(self, name, a1, a2):
        """
        SNP class to store name and alleles
        :type name: basestring
        :type a1: basestring
        :type a2: basestring
        """
        self.name = name
        self.alleles = [a1, a2]

class HapFLK(object):
    ''' A class to run an hapFLK analysis.
    '''
    def __init__(self):
        ''' Attributes of the class (not all need be specified)

        Note: Data arrays indices relate to the pops and sortes_snps lists.

            Input Data
            ----------
            - carte: data.Map instance of SNP info (may contain > nsnp)
            - sorted_snps: list of SNP objects ordered by position in carte
            - alleles: dict of SNP alleles (key=SNP names)
            - pops: list of populations

            Model Data
            -----------
            - outgroup: name of the outgroup used to build kinship. Usually not in pops
            - reynolds: matrix of reynolds distances between pop
            - hzy: heterozygosity of pops
            - kinship: Kinship matrix between pops
            - frq: SNP allele freq in pops
            - K : number of clusters
            - nfit : number of EM fits 
            - kfrq: haplotype cluster frequencies

            Test Data
            ---------
            - flk: FLK values
            - pzero: estimated ancestral allele frequencies
            - pval_FLK: theoretical p-value of FLK
            - hapflk: hapFLK values
            - eigvec: Eigen Vectors of kinship matrix
            - eigval: Eigen Values of kinship matrix
            - eigen_FLK: Spectral decomposition of FLK test
            - eigen_hapFLK: Spectral decomposition of hapFLK test
        '''
        ## SNP info
        self.carte=None
        self.sorted_snps=None
        self.alleles=None
        ## Model data
        self.pops=None
        self.outgroup=None
        self.reynolds=None
        self.hzy=None
        self.kinship=None
        self.frq=None
        self.K=-1
        self.nfit=0
        self.kfrq=None
        ## Test data
        self.flk=None
        self.pzero=None
        self.hapflk=None
        self.eigvec=None
        self.eigval=None
        self.flk_eigen=None
        self.hapflk_eigen=None

    @property
    def nsnp(self):
        return len(self.sorted_snps)

    @property
    def npop(self):
        return len(self.pops)
    
    @classmethod
    def from_db_file(cls,filename):
        '''Create Instance from dbfile'''
        with bz2.BZ2File(filename,'rb') as f:
            obj=pickle.load(f)
        obj.__class__=cls ## allows to load parent obj from subclasses
        return obj

    @classmethod
    def from_cmd_line(cls,opts):
        '''Create instance from command line arguments'''
        obj=cls()

        data=obj.read_input(opts)
        obj.model_setup(data,opts)
        obj.write_frq(opts.prefix)
        
        if opts.K>0:
            print("Fitting LD model (this might take a while)")
            fphpars=obj.run_fastphase(data,opts)
            print("Computing haplotype cluster frequencies")
            obj.calc_kfrq(data,opts,fphpars)
            if opts.kfrq:
                obj.write_kfrq(opts.prefix)
            obj.nfit,obj.K,np,ns=obj.kfrq.shape
        else:
            obj.kfrq=None
            obj.nfit=0
            obj.K=0
        return obj
        

    @staticmethod
    def populate_parser(parser):
        parser.add_argument('-p','--prefix',dest='prefix',help='prefix for output files',default='hapflk')
        parser.add_argument('--ncpu',metavar='N',help='Use N processors when possible',default=cpu_count(),type=int)
        parser.add_argument('--eigen',help='Perform eigen decomposition of tests',default=False,action='store_true')
        parser.add_argument('--reynolds',help='Force writing down Reynolds distances',default=False,action='store_true')
        parser.add_argument('--future',help=argparse.SUPPRESS,default=False,action="store_true") ## for testing future release
        parser.add_argument('--debug',help=argparse.SUPPRESS,default=False,action="store_true") ## for debug purpose
        parser.add_argument('--savedb',help=argparse.SUPPRESS,default=False,action="store_true") ## save hapflk DB
        parser.add_argument('--loaddb',dest='dbfile',help=argparse.SUPPRESS,default=None) ## load hapflk DB
        flk_opts=parser.add_argument_group('Population kinship ','Set parameters for getting the population kinship matrix')
        flk_opts.add_argument('--kinship',help='Read population kinship from file (if None, kinship is estimated)',metavar='FILE',default=None)
        flk_opts.add_argument('--reynolds-snps',dest='reysnps',type=int,help='Number of SNPs to use to estimate Reynolds distances',default=100000,metavar='L')
        flk_opts.add_argument('--outgroup',default=None,help='Use population POP as outgroup for tree rooting (if None, optimize root location)',metavar="POP")
        flk_opts.add_argument('--keep-outgroup',dest='keepOG',default=False,help='Keep outgroup in population set',action="store_true")
        flk_opts.add_argument('--covkin',default=False,help='Use covariance matrix as kinship', action='store_true')
        LD_opts=parser.add_argument_group('hapFLK and LD model','Switch on hapFLK calculations and set parameters of the LD model ')
        LD_opts.add_argument('-K',help='Set the number of clusters to K. hapFLK calculations switched off if K<0',default=-1,type=int)
        LD_opts.add_argument('--nfit',help='Set the number of model fit to use',type=int,default=10)
        LD_opts.add_argument('--phased','--inbred',help='Haplotype data provided',dest='inbred',action="store_true",default=False)
        LD_opts.add_argument('--fastLD',help=argparse.SUPPRESS,default=False,action='store_true') ## Fit LD model on "haploidized" individuals
        LD_opts.add_argument('--kfrq',dest='kfrq',help='Write Cluster frequencies (Big files)',action="store_true",default=False)
        LD_opts.add_argument('--write-params',dest='wparams',help=argparse.SUPPRESS,default=False,action='store_true')
        LD_opts.add_argument('--legacy', dest='legacy',help='Use Legacy fastPHASE', default=False, action='store_true')
        parser.add_argument('--annot',help='Shortcut for --eigen --reynolds --kfrq',default=False,action='store_true')
        parser.add_argument('--tree',help=argparse.SUPPRESS,default=False,action='store_true')

    @classmethod
    def get_opts(cls,args):
        ''' Parse arguments from a command line and returns hapFLK options'''
        ## Arguments and option parser
        parser=argparse.ArgumentParser(parents=[IO.io_parser],formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        cls.populate_parser(parser)
        if len(args)<2:
            parser.print_help()
            sys.exit(0)
        opts=parser.parse_args(args[1:])
        if opts.annot:
            opts.reynolds=True
            opts.kfrq=True
            opts.eigen=True
        if opts.sfile:
            opts.inbred=True
        return opts
    
    def read_input(self,opts):
        '''Read input files
        Creates the following attributes:
        -- carte
        -- sorted_snps
        -- nsnps
        '''
        myi=IO.parseInput(opts)
        if myi is None:
            self.parser.print_help()
            sys.exit(0)
        data=myi['dataset']
        self.carte=myi['map']
        self.sorted_snps=[]
        for sname in self.carte.sort_loci(data.snp.keys()):
            self.sorted_snps.append(SNP(sname,data.snp[sname].alleles[0],data.snp[sname].alleles[1]))
            
        return data
    
    def model_setup(self,data,opts):
        ''' Sets up FLK model. Creates the following attributes:
        -- reynolds 
        -- hzy
        -- outgroup
        -- kinship
        -- frq
        -- pops

        '''
        ## Compute allele frequencies
        print('Computing Allele Frequencies')
        res=data.compute_pop_frq()
        freqs=res['freqs']
        pops=res['pops']
        
        ## compute Reynolds distances
        print('Computing Reynolds distances')
        if opts.reysnps>self.nsnp:
            opts.reysnps=self.nsnp
        pbinom=float(opts.reysnps)/self.nsnp
        snp_subset=np.array(np.random.binomial(1,pbinom,self.nsnp),dtype=bool)
        self.reynolds=popgen.reynolds(freqs[:,snp_subset])
        self.hzy=popgen.heterozygosity(freqs[:,snp_subset])
        self.outgroup=opts.outgroup
        ## Get kinship
        if opts.kinship:
            print('Reading Kinship Matrix')
            if not opts.keepOG:
                self.kinship=popgen.popKinship_fromFile(opts.kinship,[pop for pop in pops if pop != self.outgroup])
            else:
                self.kinship=popgen.popKinship_fromFile(opts.kinship,[pop for pop in pops])
        else:
        ## estimate kinship
            print("Computing Kinship Matrix")
            self.kinship=popgen.popKinship_new(self.reynolds,pops,self.outgroup,
                                              fprefix=opts.prefix,
                                              keep_outgroup=opts.keepOG,
                                              hzy=self.hzy,
                                              dump_tree=opts.tree)
        ## store eigen decomp of kinship matrix
        self.eigval,self.eigvec=np.linalg.eigh(self.kinship)
        ## forget about the outgroup if we keep it
        if opts.keepOG:
            ## override results / initial value
            self.outgroup='Iwantmyoutgroupback'
        ## boolean array indicating kept populations
        keep_pop=np.array([x!=self.outgroup for x in pops],dtype=bool)
        self.frq=freqs[keep_pop,]
        self.pops=tuple([x for x in pops if x!=self.outgroup])
        if opts.covkin:
            print('Computing Covariance Matrix')
            ff=popgen.FLK_test(self.kinship)
            ## calculate hat(p0)
            pzero=np.array([ff.eigen_contrib(self.frq[:,s])[0] for s in range(self.nsnp)])
            sub=(pzero>0.05)&(pzero<0.95)
            ## calculate empirical covariance matrix
            myf=self.frq[:,sub]
            m=pzero[sub]
            s=np.sqrt(m*(1-m))
            myfstd=(myf-m)/s
            covkin=np.cov(myfstd,rowvar=True)
            ## Add a small term on the diagonal to ensure stable inverse
            val=np.trace(covkin)/self.npop
            assert (val >0)
            covkin=covkin+np.diag(np.ones(self.npop)*0.01*val)
            with open(opts.prefix+'_cov.txt','w') as fcov:
                for i,pi in enumerate(self.pops):
                    tw=[pi]
                    for j,pj in enumerate(self.pops):
                        tw.append(str(covkin[i,j]))
                    print(' '.join(tw),file=fcov)
            self.kinship=covkin
        
    def run_fastphase(self,data,opts):
        '''Estimates Scheet and Stephens LD model on data'''
        fphpars=[]
        with myfph.fastphase(self.nsnp, opts.ncpu, prfx = opts.prefix+'.fphlog') as fastphase_model:
            tohap=np.vectorize(lambda x: (x==1) and -1 or (x//2))
            sorted_snps_idx=np.array([data.snpIdx[s.name] for s in self.sorted_snps])
            for name,i in data.indivIdx.items():
                if data.indiv[name].pop==self.outgroup:
                    continue
                if opts.inbred:
                    haplo=np.array(tohap(np.array(data.Data[i,sorted_snps_idx],dtype=int)),dtype=np.int)
                    fastphase_model.addHaplotype(name,haplo)
                else:
                    fastphase_model.addGenotype(name,np.array(data.Data[i,sorted_snps_idx],dtype=int))
            if opts.legacy:
                for e in range(opts.nfit):
                    sys.stderr.write('\tEM %d / %d \r'%(e+1,opts.nfit))
                    sys.stderr.flush()
                    if opts.debug:
                        print()
                    par=fastphase_model.fit(nClus=opts.K,verbose=opts.debug, fast=opts.fastLD)
                    fphpars.append(par)
                    if opts.wparams:
                        fout=open(opts.prefix+'fph_par_'+str(e),'w')
                        par.write(fout)
                        fout.close()
            else:
                par = fastphase_model.optimfit(nClus=opts.K,verbose=opts.debug,nEM=opts.nfit, fast=opts.fastLD)
                # with open(opts.prefix+'debug_alpha_const.txt','w') as fpar:
                #     par.write(stream=fpar)
                ## update alpha
                # print('Finalizing: Updating Cluster Weights')
                # par.alpha_up = True
                # par = fastphase_model.fit(params=par,nClus=opts.K,verbose=opts.debug,nstep=50)
                # with open(opts.prefix+'debug_alpha_free.txt','w') as fpar:
                #     par.write(stream=fpar)
                opts.nfit = 1
                fphpars.append(par)
                if opts.wparams:
                        fout=open(opts.prefix+'fph_par','w')
                        par.write(fout)
                        fout.close()
            print()
        return fphpars
        
    def calc_kfrq(self,data,opts,fphpars):
        ''' Compute haplotype cluster population frequencies '''
        ## [E][K][npop x nsnp] array
        pop_cluster_freq=np.zeros((opts.nfit,opts.K,self.npop,self.nsnp),dtype=float)
        with myfph.fastphase( self.nsnp, opts.ncpu) as fastphase_model:
            tohap=np.vectorize(lambda x: (x==1) and -1 or (x/2))
            sorted_snps_idx=np.array([data.snpIdx[s.name] for s in self.sorted_snps])
            for ipop,popname in enumerate(self.pops):
                sys.stdout.write("\t %16s\r"%popname)
                sys.stdout.flush()
                pvec=data.populations[popname]
                n_pop_indiv=sum(pvec)
                for name,iind in data.indivIdx.items():
                    if pvec[iind]:
                        if opts.inbred:
                            haplo=np.array(tohap(np.array(data.Data[iind,sorted_snps_idx],dtype=int)),dtype=np.int)
                            fastphase_model.addHaplotype(name,haplo)
                        else:
                            fastphase_model.addGenotype(name,np.array(data.Data[iind,sorted_snps_idx],dtype=int))
                for ifit in range(opts.nfit):
                    imputations=fastphase_model.impute([fphpars[ifit]])
                    for nom,dat in imputations.items():
                        probZ=dat[1]
                        if opts.inbred:
                            pop_cluster_freq[ifit,:,ipop,:]+=np.transpose(probZ[0]/n_pop_indiv)
                        else:
                            pop_cluster_freq[ifit,:,ipop,:]+=np.transpose((0.5/n_pop_indiv)*(np.sum(probZ[0],axis=1)+np.sum(probZ[0],axis=2)))
                fastphase_model.flush()
        self.kfrq=pop_cluster_freq

    def run_tests(self,opts):
        if self.flk is None:
            print('Computing FLK tests')
            self.calc_flk()
            self.write_flk(opts.prefix)
            
        if self.K>0:
            if self.hapflk is None:
                print("Computing hapFLK tests")
                self.calc_hapflk()
                self.write_hapflk(opts.prefix)
        return
    def calc_flk(self):
        '''Computes FLK tests'''
        ## Compute single SNP FLK tests
        myFLK=popgen.FLK_test(self.kinship)
        myFLK_res=np.apply_along_axis(myFLK.eval_flk,0,self.frq)
        self.pzero=myFLK_res[0,]
        self.flk=myFLK_res[1,]
        self.pval_FLK=myFLK_res[2,]
        self.eigen_FLK=np.power(myFLK_res[3:,],2)
        return
    
    def calc_hapflk(self):
        ''' Compute hapFLK tests '''
        myFLK=popgen.FLK_test(self.kinship,diallelic=False)
        self.hapflk=np.zeros(self.nsnp,dtype=float)
        self.eigen_hapFLK=np.zeros((self.npop,self.nsnp),dtype=float)
        for e in range(self.nfit):
            for k in range(self.K):
                myFLK_res=np.apply_along_axis(myFLK.eval_flk,0,self.kfrq[e,k,])
                self.hapflk+=myFLK_res[1,]
                self.eigen_hapFLK+=np.power(myFLK_res[3:,],2)
        self.hapflk/=self.nfit
        self.eigen_hapFLK/=self.nfit
        return

    def calc_hapflk_ng(self):
        pass
    
    def savedb(self,prefix):
        ''' Save (pickle) instance of the class in a 'prefix'.db file
            NB: The instance can be loaded (unpickled) using the class method from_db_file.
        '''
        dbfile=bz2.BZ2File(prefix+'.db','wb')
        pickle.dump(self,dbfile,protocol=2)
        dbfile.close()

    def write_frq(self,prefix):
        ''' Write allele frequencies to 'prefix'.frq'''
        assert(self.frq is not None)
        with open(prefix+'.frq','w') as fout:
            print('rs','chr','pos','all_ref','all_alt',' '.join(self.pops),file=fout)
            for sidx,s in enumerate(self.sorted_snps):
                spos=self.carte.position(s.name)
                tw=[s.name,spos[0],int(spos[2]),s.alleles[1],s.alleles[0]]
                for ip,nom in enumerate(self.pops):
                    tw.append(self.frq[ip,sidx])
                print(*tw,file=fout)

    def write_kfrq(self,prefix):
        if self.kfrq is None:
            pass
        nfit,nclus,npop,nsnp = self.kfrq.shape
        for ifit in range(nfit):
            with bz2.open(prefix+'.fit_'+str(ifit)+'.bz2','wt') as fout:
                print('pop','locus','position','cluster','prob',file=fout)
                for ipop in range(npop):
                    for i,s in enumerate(self.sorted_snps):
                        spos=self.carte.position(s.name)
                        for ik in range(nclus):
                            print(self.pops[ipop],s.name,int(spos[2]),ik,self.kfrq[ifit,ik,ipop,i], file=fout)
            
    def write_flk(self,prefix):
        ''' Write FLK results to 'prefix'.flk'''
        assert(self.flk is not None)
        with open(prefix+'.flk','w') as fout:
            print('rs','chr','pos','pzero','flk','pvalue',file=fout)
            for sidx,s in enumerate(self.sorted_snps):
                spos=self.carte.position(s.name)
                tw=[s.name,spos[0],int(spos[2]),self.pzero[sidx],self.flk[sidx],self.pval_FLK[sidx]]
                print(*tw,file=fout)

    def write_hapflk(self,prefix):
        ''' Write hapFLK results to 'prefix'.hapflk'''
        assert(self.flk is not None)
        with open(prefix+'.hapflk','w') as fout:
            print('rs','chr','pos','hapflk','K',file=fout)
            for sidx,s in enumerate(self.sorted_snps):
                spos=self.carte.position(s.name)
                tw=[s.name,spos[0],int(spos[2]),self.hapflk[sidx],self.kfrq.shape[1]]
                print(*tw,file=fout)

    
 



#!python
#cython: language_level=3, boundscheck=False
'''
Module implementing the LD model of Scheet and Stephens (2006)
originally implemented in Paul Scheet fastPHASE software.
Code is largely inspired by Yongtao Guan Bimbam Code and as such
is released under the GNU General Public Licence version 2

Multithread version
'''
cimport cython
## NUMPY import in cython
import numpy as np
from scipy.optimize import linear_sum_assignment
from libc.math cimport log
cimport numpy as np
np.import_array()
from multiprocessing import Pool, cpu_count
import sys
from functools import reduce
from operator import add


class fitInData():
    def __init__(self,type,alpha,theta,rho,data,u2z):
        self.type=type
        self.alpha=alpha
        self.theta=theta
        self.rho=rho
        self.data=data
        self.up2pz=u2z

class fitOutData():
    def __init__(self,logLike,top,bot,jmk,val):
        self.logLike=logLike
        self.top=top
        self.bot=bot
        self.jmk=jmk
        self.val=val

class imputeInData():
    def __init__(self,type,parList,nLoc,name,data):
        self.type=type
        self.parList=parList
        self.name=name
        self.nLoc=nLoc
        self.data=data
        
class imputeOutData():
    def __init__(self,pgeno,name,pZ,path=None):
        self.name=name
        self.pgeno=pgeno
        self.pZ=pZ
        self.path=path

class modParams():
    '''
    A class for fastphase model parameters.
    Dimensions:
    -- number of loci: N
    -- number of clusters: K
    Parameters:
    -- theta (N x K): allele frequencies in clusters at each locus
    -- alpha (N x K): cluster weights at each locus
    -- rho (N x 1): jump probabilities in each interval
    '''
    def __init__(self,nLoc,nClus,rhomin=1e-6, alpha_up = True, theta_up = True):
        self.nLoc=nLoc
        self.nClus=nClus
        self.theta=0.98*np.random.random((nLoc,nClus))+0.01 # avoid bounds 0 and 1
        self.rho=np.ones((nLoc,1))/1000
        ##self.alpha=np.random.mtrand.dirichlet(np.ones(nClus),nLoc) # make sure sum(alpha_is)=1
        self.alpha=1.0/nClus*np.ones((nLoc,nClus))
        self.rhomin=rhomin
        self.alpha_up = alpha_up
        self.theta_up = theta_up
        self.loglike = 0
    def initUpdate(self):
        self.top=np.zeros((self.nLoc,self.nClus))
        self.bot=np.zeros((self.nLoc,self.nClus))
        self.jmk=np.zeros((self.nLoc,self.nClus))
        self.jm=np.zeros((self.nLoc,1))
        self.nhap=0.0
    def addIndivFit(self,t,b,j,nhap):
        self.top += t
        self.bot += b
        self.jmk += j
        self.jm  += np.reshape(np.sum(j,axis=1),(self.nLoc,1))
        self.nhap+=nhap
    def update(self):
        ''' Update parameters using top,bot,jmk jm probabilities'''
        ## rho
        ##self.rho=self.jm/self.nhap
        for i in range(self.nLoc):
            self.rho[i,0]=self.jm[i,0]/self.nhap
            if self.rho[i,0]<self.rhomin:
                self.rho[i,0]=self.rhomin
            elif self.rho[i,0]>(1-self.rhomin):
                self.rho[i,0]=1-self.rhomin
        ## alpha
        #self.alpha=self.jmk/self.jm
        if self.alpha_up:
            for i in range(self.nLoc):
                for j in range(self.nClus):
                    self.alpha[i,j]=self.jmk[i,j]/self.jm[i,0]
                    if self.alpha[i,j]>=0.999:
                        self.alpha[i,j]=0.999
                    elif self.alpha[i,j]<0.001:
                        self.alpha[i,j]=0.001
                self.alpha[i,:] /= np.sum(self.alpha[i,:])
        ## theta
        if self.theta_up:
            self.theta=self.top/self.bot
            for i in range(self.nLoc):
                for j in range(self.nClus):
                    if self.theta[i,j]>0.999:
                        self.theta[i,j]=0.999
                    elif self.theta[i,j]<0.001:
                        self.theta[i,j]=0.001
    def write(self,stream=sys.stdout):
        print("snp", *["t"+str(i) for i in range(self.nClus)], "rho", *["a"+str(i) for i in range(self.nClus)], file=stream)
        for i in range(self.nLoc):
            print(i, *[np.round( self.theta[i,k], 3) for k in range(self.nClus)], np.round( self.rho[i,0], 7), *[np.round( self.alpha[i,k], 3) for k in range(self.nClus)], file=stream)

_tohap = np.vectorize(lambda x: (x==1) and -1 or (x/2))
    
class fastphase():
    '''
    A class to manipulate and control a fastphase model (Scheet and Stephens, 2006)
    Initialized with a problem size = number of loci

    Usage : 
    with fastphase(nloc, nproc) as fph:
         ... do stuff ...
    '''
    def __init__(self, nLoci, nproc = cpu_count(),prfx=None):
        assert nLoci>0
        self.nLoci=nLoci
        self.haplotypes={}
        self.genotypes={}
        self.nproc = nproc
        self.pool = None
        self.prfx = prfx
        
    def __enter__(self):
        self.pool = Pool ( self.nproc)
        if self.prfx is None:
            self.flog = sys.stdout
        else:
            self.flog = open(self.prfx, 'w')
        return self

    def __exit__(self,*args):
        self.pool.terminate()
        if self.prfx is not None:
            self.flog.close()

    def flush(self):
        '''
        remove data 
        '''
        self.haplotypes={}
        self.genotypes={}
    
    def addHaplotype(self,ID,hap,missing=-1):
        '''
        Add an haplotype to the model observations.
        hap is a numpy array of shape (1,nLoci).
        Values must be 0,1 or missing
        '''
        try:
            assert hap.shape[0]==self.nLoci
            self.haplotypes[ID]=hap
        except AssertionError:
            print("Wrong Haplotype Size:",hap.shape[0],"is not",self.nLoci)
            raise
    def addGenotype(self,ID,gen,missing=-1):
        '''
        Add a genotype to the model observations.
        gen is a numpy array of shape (1,nLoci).
        Values must be 0,1,2 or missing
        '''
        try:
            assert gen.shape[0]==self.nLoci
            self.genotypes[ID]=gen
        except AssertionError:
            print("Wrong Genotype Size:",gen.shape[0],"is not",self.nLoci)
            raise

    @staticmethod
    def gen2hap(gen):
        return np.array( _tohap(np.array(gen, dtype=int)), dtype=int)
    
    def fit(self,nClus=20,nstep=20,params=None,verbose=False,rhomin=1e-6, alpha_up = True, theta_up = True, fast=False):
        '''
        Fit the model on observations with nCLus clusters using nstep EM iterations
        Multithread version.
        '''
        try:
            assert self.pool is not None
        except AssertionError:
            print('Usage :\n\t with fastphase(nloc, nproc) as fph: \n ...')
            raise
            
        if params:
            par=params
            par.alpha_up = alpha_up
            par.theta_up = theta_up
        else:
            par=modParams(self.nLoci,nClus,rhomin, alpha_up, theta_up)
        if verbose:
            print( 'Fitting fastphase model',file=self.flog)
            print( '# clusters ',nClus, file=self.flog)
            print( '# threads ', self.nproc, file=self.flog)
            print( '# Loci', self.nLoci, file=self.flog)
            print( '# Haplotypes',len(self.haplotypes), file=self.flog)
            print( '# Genotypes', len(self.genotypes), file=self.flog)
        old_log_like=1
  
        for iEM in range(nstep):
   
            log_like=0.0
            par.initUpdate()

            tasks =  [ fitInData( 'haplo', par.alpha,par.theta,par.rho,hap,0) for hap in  self.haplotypes.values()]
            if fast:
                tasks += [ fitInData( 'haplo', par.alpha,par.theta,par.rho,fastphase.gen2hap(gen),0) for gen in  self.genotypes.values()]
            else:
                tasks += [ fitInData( 'geno', par.alpha,par.theta,par.rho,gen,0) for gen in  self.genotypes.values()]
            results = self.pool.map( fitter, tasks)

            for item in results:
                par.addIndivFit(item.top,item.bot,item.jmk,item.val)
                log_like += item.logLike
            if verbose:
                print( iEM, log_like, file=self.flog)
                self.flog.flush()
            par.update()
            par.loglike=log_like
        
        return par

    def optimfit(self, nClus = 20, nstep = 10, params=None, verbose = False, rhomin=1e-6, alpha_up = False, fast=False, nEM = 10, niter=5):
        liktraj = []

        print("=== OPTIMFIT RUN ===",file=self.flog)
        ## Fit nEM EM to find an initial start
        if verbose:
            print("*** Init EM", 1,file=self.flog)
        nbest = 0
        curpar = self.fit( nClus = nClus, nstep = nstep, verbose = True, alpha_up = alpha_up, fast=fast)
        liktraj.append((0, 0, curpar.loglike))
        for n in range(1, nEM):
            if verbose:
                print("*** Init EM", n+1,file=self.flog)
            
            par = self.fit( nClus = nClus, nstep = nstep,verbose = True, alpha_up = alpha_up, fast=fast)
            liktraj.append((n, 0, par.loglike))
            if par.loglike > curpar.loglike:
                nbest = n
                curpar = par
        init_par = curpar
        ## Improve it
        print('Initial Viterbi',file=self.flog)
        self.flog.flush()
        imp = self.viterbi([curpar])
        
        for it in range(niter):
            if verbose:
                print("*** Iter", it+1,file=self.flog)
                self.flog.flush()
            ## calculate costs
            if verbose:
                print("Calculating Costs",file=self.flog)
                self.flog.flush()
            ##cost_mat = np.zeros( (self.nLoci-1, nClus, nClus), dtype=np.float)
            cost_mat = [ np.zeros( (nClus,nClus), dtype = np.float) for i in range(self.nLoci-1)]
            ### Genotypes
            gen_dat = ( ( imp[geno][0][ijump:ijump+2], nClus) for geno in self.genotypes.keys() for ijump in range(self.nLoci-1)  )
            for geno in self.genotypes.keys():
                vit = imp[geno][0]
                args = [(vit[ijump:ijump+2], nClus) for ijump in range(self.nLoci -1)]
                ## res is [ np.array(nK,nK) , ... ,np.array(nK,nK)]
                res = self.pool.map( calc_cost_matrix_geno, args, 100)
                ## " cost_mat += res "
                cost_mat = list( map( add, cost_mat, res))
            ### haplotypes
            for haplo in self.haplotypes.keys():
                vit = imp[haplo][0]
                args = [(vit[ijump:ijump+2], nClus) for ijump in range(self.nLoci -1)]
                ## res is [ np.array(nK,nK) , ... ,np.array(nK,nK)]
                res = self.pool.map( calc_cost_matrix_haplo, args, 100)
                ## " cost_mat += res "
                cost_mat = list( map( add, cost_mat, res))
                ##cost_mat += self.pool.map( calc_cost_matrix_haplo, args, 100)
            ## combine
            if verbose:
                print("Computing optimum permutations",file=self.flog)
                self.flog.flush()
            ##cost_mat = cost_mat.reshape( ( self.nLoci-1, nClus, nClus))
            ##args = [ cost_mat[i] for i in range( self.nLoci-1)]
            res = np.array( self.pool.map( linear_sum_assignment, cost_mat,  100))
            permut = res[:,1,:]
            newpar = switch_pars( curpar, permut)
            if verbose:
                print("EM with switched parameters",file=self.flog)
                self.flog.flush()
            if (it == niter-1): ## last iteration, longer and no imputation
                par_s = self.fit( nClus = nClus, nstep=max(30,nstep), verbose=True, params=newpar, alpha_up=False, fast=fast)
            else:
                par_s = self.fit( nClus = nClus, nstep=nstep, verbose=True, params=newpar, alpha_up=False, fast=fast)
                imp = self.viterbi([par_s])
            curpar = par_s
            liktraj.append((nbest, it+1, curpar.loglike))
        
        ## add alpha update
        # par_s = self.fit(nClus = nClus, nstep = max(30,nstep), verbose = True, params = par_s, alpha_up = True, theta_up = False, fast = fast)
        # liktraj.append((nbest, it+2, par_s.loglike))
        print('EM','iter','loglik',file=self.flog)
        for dat in liktraj:
            print(*dat,file=self.flog)
        return par_s

    def viterbi(self, parList):
        tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
        tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
        results = self.pool.map( viterber, tasks)
        Imputations={}
        for item in results:
            Imputations[item.name]=item.path
        return Imputations

    
        # results=[]
        # nblocks = len(tasks)//self.nproc
        # reste = len(tasks)%self.nproc
        # for ib in range(nblocks-1):
        #     results += self.pool.map( viterber, tasks[ib*nproc:(ib+1)*nproc])

        # Imputations={}
        # for item in results:
        #     Imputations[item.name]=item.path

        # ## Memory efficient approach, only create nproc tasks at a time
        # ib = 0
        # tasks = []
        # results = []

        # ##### Haplotyps
        # for name, hap in self.haplotypes.items():
        #     tasks.append(imputeInData('haplo',parList,self.nLoci,name,hap))
        #     ib += 1
        #     if ib == self.nproc:
        #         results += self.pool.map( viterber, tasks)
        #         ib = 0
        #         tasks = []
        # ## finish the job
        # if len(tasks) >0:
        #     results += self.pool.map( viterber, tasks)
        #     ib = 0
        #     tasks=[]
        # ##### Genotypes
        # for name, gen in self.genotypes.items():
        #     ib += 1
        #     tasks.append(imputeInData('geno',parList,self.nLoci,name,gen))
        #     if ib == self.nproc:
        #         results += self.pool.map( viterber, tasks)
        #         ib = 0
        #         tasks = []
        # ## finish the job
        # if len(tasks) >0:
        #     results += self.pool.map( viterber, tasks)
        #     ib = 0
        #     tasks=[]
    
    def impute(self,parList):
        tasks =  [ imputeInData('haplo',parList,self.nLoci,name,hap) for name, hap in  self.haplotypes.items()]
        tasks += [ imputeInData('geno',parList,self.nLoci,name,gen) for name, gen in self.genotypes.items()]
        results = self.pool.map( imputer, tasks)
        Imputations={}
        for item in results:
            Imputations[item.name]=(item.pgeno,item.pZ,item.path)
        return Imputations



def fitter( item):
    #type,alpha,theta,rho,data = item
    if item.type == 'haplo' :
        try:
            hLogLike,top,bot,jmk=hapCalc(item.alpha,item.theta,item.rho,item.data,0)
        except ValueError:
            print( item.data)
            raise
        res = fitOutData(hLogLike,top,bot,jmk,1)
    elif item.type == 'geno' :
        gLogLike,top,bot,jmk=genCalc(item.alpha,item.theta,item.rho,item.data,0)
        res = fitOutData(gLogLike,top,bot,jmk,2)
    return res

def viterber( item):
    #cIter=iter
    cdef int i,k
    cdef int nLoc
    cdef double x
    cdef double nsamp=100
    if item.type == 'haplo':
        path = []
        for par in item.parList:
            pth = hapViterbi( par.alpha, par.theta, par.rho, item.data)
            path.append(pth)
        res = imputeOutData(None,item.name,None,path)
    elif item.type == 'geno':
        path = []
        for par in item.parList:
            pth = genViterbi( par.alpha, par.theta, par.rho, item.data)
            path.append( pth)
        res = imputeOutData(None,item.name,None, path)
    return res

def imputer( item):
    #cIter=iter
    cdef int i,k
    cdef int nLoc
    cdef double x
    cdef double nsamp=100
    if item.type == 'haplo':
        pgeno=np.zeros(item.nLoc,dtype=np.float64)
        probZ=[]
        path = []
        x = 1.0/len(item.parList)
        for par in item.parList:
            pZ=hapCalc(par.alpha,par.theta,par.rho,item.data,1)
            for i in range(item.nLoc):
                for k in range(par.nClus):
                    pgeno[i]+=x*hap_p_all(item.data[i],pZ[i,k],par.theta[i,k])
            probZ.append(pZ)
            pth = None#hapViterbi( par.alpha, par.theta, par.rho, item.data)
            path.append(pth)
        res = imputeOutData(pgeno,item.name,probZ,path)
    elif item.type == 'geno':
        pgeno=np.zeros(item.nLoc,dtype=np.float64)
        probZ=[]
        path = []
        x = 1.0/len(item.parList)
        for par in item.parList:
            pZ=genCalc(par.alpha,par.theta,par.rho,item.data,1)
            ## Calculate mean genotypes
            for i in range(item.nLoc):
                for k1 in range(par.nClus):
                    for k2 in range(par.nClus):
                        pgeno[i]+=gen_p_geno(item.data[i],pZ[i,k1,k2],par.theta[i,k1],par.theta[i,k2])
            probZ.append(pZ)
            pth = None#genViterbi( par.alpha, par.theta, par.rho, item.data)
            path.append( pth)
        pgeno/=len(item.parList)
        res = imputeOutData(pgeno,item.name,probZ, path)
    return res

##### Cluster switch functions

cpdef np.ndarray[np.float64_t, ndim=3] calc_cost_matrix_haplo_tot( np.ndarray[np.int32_t, ndim=1] h, int nK ):
    cdef int l,nL
    nL= h.shape[0]
    cdef np.ndarray[np.float64_t, ndim=3] res = np.zeros( ( nL-1, nK, nK), dtype=np.float64)
    
    for l in range(nL -1):
        res[l, h[l], h[l+1]] -= 1
    return res

def calc_cost_matrix_haplo( args):
    vit, nK = args
    k = vit[0]
    kp = vit[1]
    cost_mat = np.zeros((nK,nK))
    cost_mat[k, kp] -= 1
    return cost_mat

cpdef np.ndarray[np.float64_t, ndim=3] calc_cost_matrix_geno_tot( np.ndarray[np.int32_t, ndim=2] g, int nK ):
    cdef int l,nL
    cdef int k1, k2, kp1, kp2
    nL= g.shape[0]
    cdef np.ndarray[np.float64_t, ndim=3] res = np.zeros( ( nL-1, nK, nK), dtype=np.float64)

    for l in range(nL -1):
        if (k1 == kp1):
            if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
                res[l, k1, kp1] -= 1
                res[l, k2, kp2] -= 1
            else: ## k1 == kp1 & k2 != kp2
                res[l, k1, kp1] -= 1 ## diagonal
                res[l, k2, kp2] -= 1
        elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
            res[l, k1, kp1] -= 1 ## diagonal
            res[l, k2, kp2] -= 1
        elif ( k1 == kp2) and (k2 != kp1):
            res[l, k1, kp2] -= 1 ## diagonal
            res[l, k2, kp1] -= 1
        elif ( k1 != kp2) and (k2 == kp1):
            res[l, k1, kp2] -= 1 ## diagonal
            res[k2, kp1] -= 1
        else:
            res[l, k1, kp1] -= 0.5
            res[l, k1, kp2] -= 0.5
            res[l, k2, kp1] -= 0.5
            res[l, k2, kp2] -= 0.5
    return res

def calc_cost_matrix_geno( args):
    vit, nK = args
    k1, k2 = vit[0]
    kp1, kp2 = vit[1]
    cost_mat = np.zeros((nK,nK))
    if (k1 == kp1):
        if (k2 == kp2): ## k1 == kp1 & k2 == kp2 , possibly k1 == k2
            cost_mat[k1, kp1] -= 1
            cost_mat[k2, kp2] -= 1
        else: ## k1 == kp1 & k2 != kp2
            cost_mat[k1, kp1] -= 1 ## diagonal
            cost_mat[k2, kp2] -= 1
    elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
        cost_mat[k1, kp1] -= 1 ## diagonal
        cost_mat[k2, kp2] -= 1
    elif ( k1 == kp2) and (k2 != kp1):
        cost_mat[k1, kp2] -= 1 ## diagonal
        cost_mat[k2, kp1] -= 1
    elif ( k1 != kp2) and (k2 == kp1):
        cost_mat[k1, kp2] -= 1 ## diagonal
        cost_mat[k2, kp1] -= 1
    else:
        cost_mat[ k1, kp1] -= 0.5
        cost_mat[ k1, kp2] -= 0.5
        cost_mat[ k2, kp1] -= 0.5
        cost_mat[ k2, kp2] -= 0.5
    return cost_mat


def switch_pars(par, permut):
    newpar = modParams(par.nLoc, par.nClus, alpha_up = par.alpha_up)
    for i in range(par.nClus):
        newpar.theta[0,i]=par.theta[0,i]
        curclus = i
        for j in range(1, par.nLoc):
            curclus =  permut[ j-1, curclus]
            newpar.theta[ j,i] = par.theta[ j, curclus]
    return newpar

############################### Calculations #################################

cdef double hap_p_all(int i, double pz, double theta):
    cdef double rez
    if i<0:
        rez=pz*theta
    else:
        rez=i*pz
    return rez

cdef double gen_p_geno(int i, double pz, double theta1, double theta2):
    cdef double rez
    ## theta1*(1-theta2)+theta2*(1-theta1)+2*theta2*theta1 == theta1+theta2
    if i<0:
        rez=pz*(theta1+theta2)
    else:
        rez=i*pz
        #rez=pz*(theta1+theta2)
    return rez


cdef double myPow10(int x):
    cdef double rez=1.0
    cdef int i
    if x==0:
        return rez
    elif x<0:
        for i in range(-x):
            rez*=0.1
    else:
        for i in range(x):
            rez*=10
    return rez


cdef int argmax( np.ndarray[ np.float64_t, ndim=1] a, int size):
    cdef int i, best_i
    cdef double v, best_v

    best_v = a[ 0 ]
    best_i = 0
    for i in range(1, size):
        if a[ i ] > best_v:
            best_i = i
            best_v = a[ i ]
    return best_i

cdef int pair2idx( int k1, int k2, int nK):
    cdef int k, l
    k = min( k1, k2)
    l = max( k1, k2)
    return  nK * k - ( k * ( k -1))//2  + ( l - k)

cdef ( int, int) idx2pair( int idx, int nK):
    cdef int k, l, nelem, curk
    nelem = 0
    for curk in range(nK):
        if ( idx - nelem) < ( nK - curk):
            l = curk + (idx - nelem)
            k = curk
            break
        else:
            nelem += nK - curk
    return (k, l)


##### Genotype Calculations 

cdef double genprG(double t1, double t2, int g):
    cdef double rez
    if g==0:
        rez=(1-t1)*(1-t2)
    elif g==1:
        rez=t1+t2-2*t1*t2
    elif g==2:
        rez=t1*t2
    else:
        rez=1
    return rez

cdef double probJ(int m,int s, double rho):
    if s==0:
        return (1-rho)*(1-rho)
    elif s==1:
        return 2*(1-rho)*rho
    elif s==2:
        return rho*rho

cpdef genViterbi( aa, tt, rr, gg):
    cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
    cdef np.ndarray[np.float64_t,ndim=2] theta = tt
    cdef np.ndarray[np.float64_t, ndim=2] rho = rr
    cdef np.ndarray[np.int_t, ndim=1] gen = gg

    ## cython declarations
    cdef int nLoc, nK, ikp,  prev_kp, npairs
    cdef int k1, k2, kp1, kp2, m
    cdef double best_v
    
    nLoc = alpha.shape[0]
    nK = alpha.shape[1]
    npairs = nK * ( nK + 1)//2
    
    cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((npairs, nLoc), dtype=np.float64)
    cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((npairs, nLoc), dtype=np.int)
    cdef np.ndarray[ np.int_t, ndim = 1] soluce = np.zeros(nLoc, dtype= np.int)
    cdef np.ndarray[ np.float64_t, ndim = 1] tempVal = np.zeros( npairs, dtype=np.float64)


    ## initialization
    for k1 in range(nK):
        ikp = pair2idx( k1, k1, nK)
        delta[ikp,0] = log(alpha[ 0, k1]) + log(alpha[ 0, k1]) +log( genprG( theta[ 0, k1], theta[ 0, k1], gen[0]))
        for k2 in range( k1+1 , nK):
            ikp = pair2idx( k1, k2, nK)
            delta[ikp,0] = log(2) + log( alpha[ 0, k1]) + log( alpha[ 0, k2]) +log( genprG( theta[ 0, k1], theta[ 0, k2], gen[0]))

    ## recursion
    for m in range(1, nLoc):
        for ikp in range(npairs):
            k1, k2 = idx2pair( ikp, nK)
            for prev_kp in range(npairs):
                kp1, kp2 = idx2pair( prev_kp, nK)
                if (k1 == kp1):
                    if (k2 == kp2): ## k1 == kp1 & k2 == kp2
                        tempVal[ prev_kp] = log(probJ(m, 0, rho[m, 0])) +  delta[prev_kp, m-1]
                    else: ## k1 == kp1 & k2 != kp2
                        tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k2]) + delta[ prev_kp, m-1]
                elif ( k2 == kp2): ## k1 != kp1 & k2 == kp2
                    tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) +  delta[ prev_kp, m-1]
                elif ( k1 == kp2) or (k2 == kp1): ## crossing
                    tempVal[ prev_kp] = log(probJ(m, 1, rho[m, 0])) + log( alpha[ m, k1]) + delta[ prev_kp, m-1]
                else: ## k1 !=kp1 & k2 != kp2
                    tempVal[ prev_kp] = log(probJ(m, 2, rho[m, 0])) + log( alpha[ m, k1]) + alpha[m, k2] +  delta[ prev_kp, m-1]
            psi[ ikp, m] = argmax(tempVal, npairs)
            delta[ ikp, m] = tempVal[ psi[ ikp, m]] + log( genprG( theta[ m, k1], theta[ m, k2], gen[m]))

    ## termination
    soluce[ nLoc - 1 ] = argmax( delta[ :, nLoc-1], npairs)
    for m in range( nLoc - 2, -1, -1):
        soluce[ m ] = psi[ soluce[ m+1 ], m+1]
    return [ idx2pair(ikp, nK) for ikp in soluce]
            
cpdef genCalc(aa,tt,rr,gg,u2p):
    cdef np.ndarray[np.float64_t, ndim=2] alpha=aa
    cdef np.ndarray[np.float64_t,ndim=2] theta=tt
    cdef np.ndarray[np.float64_t, ndim=2] rho=rr
    cdef np.ndarray[np.int_t, ndim=1] gen=gg
    cdef int up2pz=u2p
    ## implementation comment: TODO = try to minimize the memory footprint
    ## of big Nloc*nK*nK matrices

    ## cython declarations
    cdef int tScale
    cdef double dummy,t1,t2
    cdef double temp,tScaleTemp
    cdef double normC
    cdef intnLoc,nK
    cdef int k,k1,k2,m
    cdef double logLikelihood
    ## end cython declarations

    nLoc=alpha.shape[0]
    nK=alpha.shape[1]
    ##
    ## compute backward probabilities
    ##
    cdef np.ndarray[np.int_t,ndim=1] betaScale=np.zeros(nLoc,dtype=np.int)
    cdef np.ndarray[np.float64_t,ndim=2] tSumk=np.zeros((nLoc,nK),dtype=np.float64)
    cdef np.ndarray[np.float64_t,ndim=1] tDoubleSum=np.zeros(nLoc,dtype=np.float64)
    cdef np.ndarray[np.float64_t,ndim=3] mBeta=np.zeros((nLoc,nK,nK),dtype=np.float64)

    for k1 in range(nK):
        for k2 in range(nK):
            mBeta[nLoc-1,k1,k2]=1
    ## calc the marginal sum at locus M as in appendix A
    for k1 in range(nK):
        for k2 in range(nK):
            tSumk[nLoc-1,k1] += genprG(theta[nLoc-1,k1],theta[nLoc-1,k2],gen[nLoc-1])*mBeta[nLoc-1,k1,k2]*alpha[nLoc-1,k2]
        tDoubleSum[nLoc-1]+=tSumk[nLoc-1,k1]*alpha[nLoc-1,k1]
    ## recurrent calculations backward
    for m in range(nLoc-1,0,-1):
        ## these loops could be parallelized across cluster pairs #CUDA
        for k1 in range(nK):
            for k2 in range(k1,nK):
                temp=0.5*probJ(m,1,rho[m,0])*(tSumk[m,k1]+tSumk[m,k2])
                temp+=probJ(m,0,rho[m,0])*genprG(theta[m,k1],theta[m,k2],gen[m])*mBeta[m,k1,k2]
                mBeta[m-1,k1,k2]=temp+probJ(m,2,rho[m,0])*tDoubleSum[m]
                mBeta[m-1,k2,k1]=mBeta[m-1,k1,k2]
        ## marginal sum and real sum for loc m-1
        for k1 in range(nK):
            for k2 in range(nK):
                tSumk[m-1,k1]+=genprG(theta[m-1,k1],theta[m-1,k2],gen[m-1])*mBeta[m-1,k1,k2]*alpha[m-1,k2]
            tDoubleSum[m-1]+=tSumk[m-1,k1]*alpha[m-1,k1]
        tScaleTemp=np.sum(mBeta[m-1])
        ## tScaleTemp=0 a gerer <---
        if tScaleTemp<1e-20:
            tScale=20
        else:
            tScale=int(-log(tScaleTemp)/log(10))
        if tScale <=0:
            tScale=0
        betaScale[m-1]=betaScale[m]+tScale
        if tScale>0:
            dummy=myPow10(tScale)
            tDoubleSum[m-1]*=dummy
            for k1 in range(nK):
                tSumk[m-1,k1]*=dummy
                mBeta[m-1,k1,k1]*=dummy
                for k2 in range(k1+1,nK):
                    mBeta[m-1,k1,k2]*=dummy
                    mBeta[m-1,k2,k1]*=dummy
    ##
    ## compute forward probabilities
    ##
    cdef np.ndarray[np.float64_t,ndim=3] mPhi=np.zeros((nLoc,nK,nK),dtype=np.float64)
    cdef np.ndarray[np.int_t,ndim=1] phiScale=np.zeros(nLoc,dtype=np.int)
    ## at locus 0
    for k1 in range(nK):
        for k2 in range(k1,nK):
            mPhi[0,k1,k2]=alpha[0,k1]*alpha[0,k2]*genprG(theta[0,k1],theta[0,k2],gen[0])
            mPhi[0,k2,k1]=mPhi[0,k1,k2]
    ## calc the marginal sum at locus 0 (appx A)
    tDoubleSum[0]=0
    for k1 in range(nK):
        tSumk[0,k1]=0
        for k2 in range(nK):
            tSumk[0,k1]+=mPhi[0,k1,k2]
        tDoubleSum[0]+=tSumk[0,k1]
    tScale=0
    if tDoubleSum[0] != 0:
        tScale=int(-log(tDoubleSum[0])/log(10))
        if tScale <0:
            tScale=0
        phiScale[0]=tScale
        if tScale>0:
            dummy=myPow10(tScale)
            for k1 in range(nK):
                tSumk[0,k1]*=dummy
                mPhi[0,k1,k1]*=dummy
                for k2 in range(k1+1,nK):
                    mPhi[0,k1,k2]*=dummy
                    mPhi[0,k2,k1]*=dummy
    # do the reccurence
    for m in range(nLoc-1):
        ## This loop can be parallelized across cluster pairs #CUDA
        for k1 in range(nK):
            for k2 in range(k1,nK):
                temp=alpha[m+1,k1]*tSumk[m,k2]+alpha[m+1,k2]*tSumk[m,k1]
                temp*=0.5*probJ(m+1,1,rho[m+1,0])
                temp+=probJ(m+1,0,rho[m+1,0])*mPhi[m,k1,k2]
                temp+=probJ(m+1,2,rho[m+1,0])*alpha[m+1,k1]*alpha[m+1,k2]*tDoubleSum[m]
                mPhi[m+1,k1,k2]=temp*genprG(theta[m+1,k1],theta[m+1,k2],gen[m+1])
                mPhi[m+1,k2,k1]=mPhi[m+1,k1,k2]
        tDoubleSum[m+1]=0
        for k1 in range(nK):
            tSumk[m+1,k1]=0
            for k2 in range(nK):
                tSumk[m+1,k1]+=mPhi[m+1,k1,k2]
            tDoubleSum[m+1]+=tSumk[m+1,k1]
        tScale=0
        if tDoubleSum[m+1]<=0:
            phiScale[m+1]=phiScale[m]
        else:
            tScale=int(-log(tDoubleSum[m+1])/log(10))
            if tScale<0:
                tScale=0
            phiScale[m+1]=phiScale[m]+tScale
            if tScale>0:
                dummy=myPow10(tScale)
                tDoubleSum[m+1]*=dummy
                for k1 in range(nK):
                    tSumk[m+1,k1]*=dummy
                    mPhi[m+1,k1,k1]*=dummy
                    for k2 in range(k1+1,nK):
                        mPhi[m+1,k1,k2]*=dummy
                        mPhi[m+1,k2,k1]*=dummy
    ## end mPhi
    logLikelihood=log(tDoubleSum[nLoc-1])/log(10)-phiScale[nLoc-1]
    ##
    ## compute Individual Contribution top,bottom,jmk
    ##
    # calc ProbZ
    cdef np.ndarray[np.float64_t,ndim=3] probZ=mPhi*mBeta
    cdef np.ndarray[np.float64_t,ndim=2] jmk=np.zeros((nLoc,nK),dtype=np.float64)
    cdef np.ndarray[np.float64_t,ndim=2] top=np.zeros((nLoc,nK),dtype=np.float64)
    cdef np.ndarray[np.float64_t,ndim=2] bot=np.zeros((nLoc,nK),dtype=np.float64)
    ## normalize
    for m in range(nLoc):
        normC=0
        for k1 in range(nK):
            normC+=probZ[m,k1,k1]
            for k2 in range(k1+1,nK):
                normC+=2*probZ[m,k1,k2]
        probZ[m]/=normC
    if up2pz>0:
        return probZ
    # calc jmk
    for k1 in range(nK):
        ##jmk[0,k1]=2*alpha[0,k1]
        jmk[0, k1] = probZ[0,k1,k1] 
        for k2 in range(nK):
            jmk[0, k1] += probZ[0, k2, k1] 
            
    for m in range(1,nLoc):
        dummy = myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])
        for k in range(nK):
            for k1 in range(nK):
                temp = tSumk[ m-1, k1] * probJ(m,1,rho[m,0])
                temp += 2 * probJ( m, 2, rho[m,0]) * tDoubleSum[m-1] * alpha[m,k1]
                jmk[m,k] += temp * genprG( theta[m,k], theta[m,k1], gen[m]) * mBeta[m,k,k1]
            jmk[m,k] *= alpha[m,k]
            jmk[m,k] /= tDoubleSum[nLoc-1]
            jmk[m,k] /= dummy
    # calc top,bottom
    for m in range(nLoc):
        ## bottom
        for k in range(nK):
            bot[m,k]+=probZ[m,k,k]
            for k1 in range(nK):
                bot[m,k]+=probZ[m,k1,k]
        ## top
        if gen[m]==0:
            for k in range(nK):
                top[m,k]=0
        elif gen[m]==1:
            for k in range(nK):
                for k1 in range(nK):
                    t1=theta[m,k]*(1-theta[m,k1])
                    t2=t1+theta[m,k1]*(1-theta[m,k])
                    if (k == k1):
                        top[m,k] += 2*probZ[m,k,k1]*t1/t2
                    else:
                        top[m,k] += probZ[m,k,k1]*t1/t2
        elif gen[m]==2:
            for k in range(nK):
                top[m,k] += probZ[m,k,k]
                for k1 in range(nK):
                    top[m,k] += probZ[m,k,k1]
        else:
            for k in range(nK):
                top[m,k]=0
                bot[m,k]=0
    return logLikelihood,top,bot,jmk

#### Haplotype Calculations
    
cdef double happrG(double t,int s):
    if s==0:
        return 1-t
    elif s==1:
        return t
    else:
        return 1
  
cpdef hapViterbi( aa, tt, rr, hh):
    cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
    cdef np.ndarray[np.float64_t,ndim=2] theta = tt
    cdef np.ndarray[np.float64_t, ndim=2] rho = rr
    cdef np.ndarray[np.int_t, ndim=1] hap = hh

    ## cython declarations
    cdef int nLoc, nK
    cdef int k, prev_k, best_k, m
    cdef double best_v
    nLoc = alpha.shape[0]
    nK = alpha.shape[1]
    cdef np.ndarray[ np.float64_t, ndim = 2 ] delta = np.zeros((nK, nLoc), dtype=np.float64)
    cdef np.ndarray[ np.int_t, ndim = 2 ] psi = np.zeros((nK, nLoc), dtype=np.int)
    cdef np.ndarray[ np.int_t, ndim = 1] soluce = np.empty(nLoc, dtype= np.int)
    cdef np.ndarray[ np.float64_t, ndim = 1] tempVal = np.zeros( nK, dtype=np.float64)
    
    ## initialization
    for k in range(nK):
        delta[k,0] = log(alpha[ 0, k]) + log(happrG( theta[ 0, k], hap[0]))

    ## recursion
    for m in range(1, nLoc):
        for k in range(nK):
            for prev_k in range(nK):
                ## jump
                tempVal[ prev_k] = log(rho[ m, 0])+log( alpha[ m, k])
                if k == prev_k:
                    ## no jump
                    tempVal[ prev_k] = log( rho[ m, 0]*alpha[ m, k] + (1 - rho[ m, 0]))
                tempVal[ prev_k] += delta[prev_k, m-1]
                psi[ k, m] = argmax(tempVal, nK)
                delta[ k, m] = tempVal[ psi[ k, m]] +log( happrG( theta[ m, k], hap[ m]))
    
    ## termination
    soluce[ nLoc - 1 ] = argmax( delta[ :, nLoc-1], nK)
    for m in range( nLoc - 2, -1, -1):
        soluce[ m ] = psi[ soluce[ m+1 ], m+1]
    return soluce
            
cpdef hapCalc(aa,tt,rr,hh,u2p):
    cdef np.ndarray[np.float64_t, ndim=2] alpha = aa
    cdef np.ndarray[np.float64_t,ndim=2] theta = tt
    cdef np.ndarray[np.float64_t, ndim=2] rho = rr
    cdef np.ndarray[np.int_t, ndim=1] hap = hh
    cdef int up2pz=u2p
    ## cython declarations
    cdef int nLoc,nK,tScale
    cdef int k,m
    cdef double dummy,tScaleTemp,temp
    cdef double normC
    cdef double logLikelihood
    ## end cython declarations
    nLoc=alpha.shape[0]
    nK=alpha.shape[1]
    ##
    ## compute backward probabilities
    ##
    cdef np.ndarray[np.int_t,ndim=1] betaScale=np.zeros(nLoc,dtype=np.int)
    cdef np.ndarray[np.float64_t,ndim=1] tSum=np.zeros(nLoc,dtype=np.float64)
    cdef np.ndarray[np.float64_t,ndim=2] mBeta=np.zeros((nLoc,nK),dtype=np.float64)

    for k in range(nK):
        mBeta[nLoc-1,k]=1

    for k in range(nK):
        tSum[nLoc-1] += happrG(theta[nLoc-1,k],hap[nLoc-1])*alpha[nLoc-1,k]
    for m in range(nLoc-1,0,-1):
        tScaleTemp=0
        ## this loop can be parallelized across clusters #CUDA
        for k in range(nK):
            temp=(1.0-rho[m,0])*happrG(theta[m,k],hap[m])*mBeta[m,k]
            mBeta[m-1,k]=temp+rho[m,0]*tSum[m]
            tSum[m-1]+=happrG(theta[m-1,k],hap[m-1])*mBeta[m-1,k]*alpha[m-1,k]
            tScaleTemp+=mBeta[m-1,k]
        # if np.isnan(tScaleTemp):
        #     print mBeta[m-1]
        #     print mBeta[m]
        #     print alpha[m-1]
        #     print theta[m]
        #     print rho[m]
        #     print hap[m]
        #     print 'BLA'
        tScale=int(-log(tScaleTemp)/log(10))
        if tScale<0:
            tScale=0
        betaScale[m-1]=betaScale[m]+tScale
        if tScale >0:
            dummy=myPow10(tScale)
            tSum[m-1]*=dummy
            for k in range(nK):
                mBeta[m-1,k]*=dummy
    ##
    ## compute forward probabilities
    ##
    cdef np.ndarray[np.float64_t,ndim=2] mPhi=np.zeros((nLoc,nK),dtype=np.float64)
    cdef np.ndarray[np.int_t,ndim=1] phiScale=np.zeros(nLoc,dtype=np.int)
    for k in range(nK):
        mPhi[0,k]=alpha[0,k]*happrG(theta[0,k],hap[0])
    ## calc the marginal sum at locus 0 (appx A)
    tSum[0]=0
    for k in range(nK):
        tSum[0]+=mPhi[0,k]
    ## calc Phi 
    for m in range(nLoc-1):
        tSum[m+1]=0
        ## this loop could be parallelized across clusters #CUDA
        for k in range(nK):
            temp=(1-rho[m+1,0])*mPhi[m,k]+rho[m+1,0]*alpha[m+1,k]*tSum[m]
            mPhi[m+1,k]=temp*happrG(theta[m+1,k],hap[m+1])
            tSum[m+1]+=mPhi[m+1,k]
        tScale=0
        if tSum[m+1] < 0:
            phiScale[m+1]=phiScale[m]
        else:
            tScale=int(-log(tSum[m+1])/log(10))
        if tScale < 0:
            tScale=0
        phiScale[m+1]=phiScale[m]+tScale
        if tScale > 0:
            dummy=myPow10(tScale)
            tSum[m+1]*=dummy
            for k in range(nK):
                mPhi[m+1,k]*=dummy
    ## end calc Phi
    logLikelihood=log(tSum[nLoc-1])/log(10)-phiScale[nLoc-1]
    ##
    ## compute individual contributions top,bottom,jmk
    ##
    # compute probZ see appx A at the end
    cdef np.ndarray[np.float64_t,ndim=2] probZ=mPhi*mBeta
    cdef np.ndarray[np.float64_t,ndim=2] jmk=np.zeros((nLoc,nK),dtype=np.float64)
    cdef np.ndarray[np.float64_t,ndim=2] top=np.zeros((nLoc,nK),dtype=np.float64)
    cdef np.ndarray[np.float64_t,ndim=2] bot=np.zeros((nLoc,nK),dtype=np.float64)
    
    for m in range(nLoc):
        normC=0
        for k in range(nK):
            normC+=probZ[m,k]
        for k in range(nK):
            probZ[m,k]/=normC
    if up2pz:
        return probZ
    ## compute expected jum prob for each interval (appx C)
    # locus 0
    for k in range(nK):
        ##jmk[0,k]=alpha[0,k]
        jmk[0, k] = probZ[ 0, k]
    for m in range(1,nLoc):
        dummy=myPow10(phiScale[m-1]+betaScale[m]-phiScale[nLoc-1])
        for k in range(nK):
            jmk[m,k]  = tSum[m-1]*rho[m,0]*happrG(theta[m,k],hap[m])*mBeta[m,k]
            jmk[m,k] *= alpha[m,k]
            jmk[m,k] /= tSum[nLoc-1]
            jmk[m,k] /= dummy
            # if not np.isfinite(jmk[m,k]):
            #     print tSum[m-1],rho[m,0],theta[m,k],hap[m],mBeta[m,k]
            #     print alpha[m,k]
            #     print tSum[nLoc-1]
            #     print dummy,phiScale[m-1],betaScale[m],phiScale[nLoc-1]
            #     raise ValueError
    # calc thetablock and its inner product with probZ (appx C.)
    for m in range(nLoc):
        if hap[m] == 0:
            for k in range(nK):
                bot[m,k]=probZ[m,k]
        elif hap[m]==1:
            for k in range(nK):
                top[m,k]=probZ[m,k]
                bot[m,k]=probZ[m,k]
    return logLikelihood,top,bot,jmk





import os
from setuptools import setup, find_packages
from setuptools.extension import Extension
from Cython.Build import cythonize

import numpy as np

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

ext_modules=[
    Extension('hapflk.fastphase',
              sources = ["fastphase/fastphase.pyx"],
##              libraries=['m'],
                  include_dirs=[np.get_include()]),
    Extension('hapflk._pgenlib',
               sources = ['pgenlib/_pgenlib.pyx', 'pgenlib/pgenlib_python_support.cpp', 'pgenlib/pgenlib_internal.cpp', 'pgenlib/plink2_base.cpp'],
              language = "c++",
              # do not compile as c++11, since cython doesn't yet support
              # overload of uint32_t operator
              # extra_compile_args = ["-std=c++11", "-Wno-unused-function"],
              # extra_link_args = ["-std=c++11"],
              extra_compile_args = ["-std=c++98", "-Wno-unused-function"],
              extra_link_args = ["-std=c++98"],
              include_dirs = [np.get_include()]
              )
   ]
                  
setup(
    name='hapflk',
    version='2.0-dev1',
    description='haplotype-based test for differentiation in multiple populations',
    long_description=read('README'),
    license = "GPL v3",
    author='Bertrand Servin',
    author_email="bertrand.servin@inra.fr",
    url='https://forge-dga.jouy.inra.fr/projects/hapflk',
    packages=['hapflk'],
    ext_modules = cythonize(ext_modules),
    setup_requires= ['numpy','Cython'],
    install_requires = [
        'numpy',
        'scipy',
        'cython'],
    scripts=["bin/hapflk", "bin/hapflkadapt", "bin/flkpoptree", "bin/poolflkadapt", "bin/poolflkannot", "bin/flkfreq"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        ]
)



'''
Module for GenomeWide Association Studies analysis
'''
from numpy import int16,vectorize

missing=int16(-1)

def _complete_cases(x):
    return x!=missing

complete_cases=vectorize(_complete_cases)

def _is_na(x):
    return x==missing

is_na=vectorize(_is_na)



import sys
import os
import pickle
import tempfile
import numpy as np
from scipy.stats import chi2,norm
from hapflk import nj
import warnings

debug=False
if not debug:
    np.seterr(all='ignore')
    warnings.filterwarnings('ignore',category=FutureWarning)

def reynolds_multi_allelic(Mf,nloc,decompose=False):
    '''
    Compute Reynolds distances from multiallelic loci

    Parameters:
    -----------
    
    Mf : numpy array of allele frequencies in populations
         rows are populations
         columns are allele frequencies, by loci
    nloc : number of loci

    Returns:
    --------

    dist : numpy array (n x n) of reynolds genetic distances
    numerator    : numerator of the Reynolds Distance
    denominator  : denominator of the Reynolds Distance
    '''
    npop=Mf.shape[0]
    A=np.dot(Mf,Mf.T)
    dist=np.zeros((npop,npop))
    if decompose:
        numerator=np.zeros((npop,npop))
        denominator=np.zeros((npop,npop))
    else:
        numerator=None
        denominator=None
    for i in range(npop-1):
        for j in range(i+1,npop):
            dist[i,j]=0.5*(A[i,i]+A[j,j]-2*A[i,j])/(nloc-A[i,j])
            if decompose:
                numerator[i,j]=A[i,i]+A[j,j]-2*A[i,j]
                denominator[i,j]=2*(nloc-A[i,j])
    dist=dist+dist.T
    numerator=numerator+numerator.T
    denominator=denominator+denominator.T
    return dist,numerator,denominator
   
def reynolds(Mf):
    '''
    Compute Reynolds distances from SNP allele frequencies

    Parameters:
    ----------------

    Mf : numpy array of allele frequencies in populations
          rows are populations (n), columns are markers (p).

    Returns:
    -----------

    dist : numpy array (n x n) of reynolds genetic distances
    '''
    npop,nloc=Mf.shape
    dist=np.zeros((npop,npop))
    A=np.dot(Mf,Mf.T)+np.dot((1-Mf),(1-Mf).T)
    for i in range(npop-1):
        for j in range(i+1,npop):
            dist[i,j]=0.5*(A[i,i]+A[j,j]-2*A[i,j])/(nloc-A[i,j])
    dist=dist+dist.T
    return dist

def heterozygosity(Mf):
    '''
    Compute heterozygosity from SNP allele frequencies
    Parameters:
    ----------------

    Mf : numpy array of allele frequencies in populations
          rows are populations (n), columns are markers (p).

    Returns:
    -----------

    hzy : numpy array (n x 1) of mean heterozygosities
    '''
    npop,nloc=Mf.shape
    hzy_func = lambda x: 2*x*(1-x)
    return np.average(hzy_func(Mf),axis=1)

def popKinship_fromFile(file_name,popnames):
    '''
    Read population kinship matrix from file

    File format is :

    Pop1 a11 a12 ...
    Pop2 a21 a22 ....

    where PopN is the population name (must be in popnames)
    aij is the kinship coefficient between pop i and j

    Parameters:
    -----------
    file_name : name of the file to read kinship from
    popnames : name of the pop to find in the file (order is preserved)

    See Also:
    ---------
    popKinship for estimating the kinship matrix
    '''

    data=[x.split() for x in open(file_name).readlines()]
    ##popnames2=popnames[:]
    # if outgroup is not None:
    #     popnames2.remove(outgroup)
    ordre=[popnames.index(x[0]) for x in data]
    if len(ordre) != len(data):
        print( 'Not Enough populations in kinship file !')
        raise ValueError
    FF=np.zeros((len(popnames),len(popnames)),dtype=float)
    for i in range(len(popnames)):
        ix=ordre[i]
        for j in range(len(popnames)):
            jx=ordre[j]
            FF[ix,jx]=float(data[i][j+1])
    return FF

def popKinship_new(D,popnames,outgroup=None,fprefix=None,keep_outgroup=False,hzy=None,dump_tree=True):
    '''
    Compute kinship matrix between populations

    Parameters:
    ---------------

    D : Reynolds distances between populations
    popnames : names of populations
    outgroup : population to use as outgroup for rooting (if None, use hzy (should give good results) 
               or do midpoint rooting (not implemented))
    fprefix : if set, save the temporary files with names built from fprefix
    keep_outgroup : do not drop the population used for rooting the tree
    hzy : Optimize root finding on observed heterozygosities vector hzy

    Returns:
    -----------

    Fij : Kinship matrix between populations
    fprefix_{ }.txt : if fprefix was set

    See Also:
    ------------

    popgen.reynolds to compute reynolds distances between populations
    popgen.heterozygosity to compute heterozygosities of populations
    '''
    if outgroup is not None:
        try:
            assert outgroup in popnames
        except AssertionError:
            print( "Outgroup not found in file",outgroup)
            print( *popnames)
            sys.exit(1)
    if not fprefix:
        buf,rey_file_name=tempfile.mkstemp(prefix='reynolds')
        buf,fij_file_name=tempfile.mkstemp(prefix='fij',suffix='.txt')
        buf,tree_file_name=tempfile.mkstemp(prefix='tree',suffix='.txt')
        buf,treedump_file_name=tempfile.mkstemp(prefix='treedump',suffix='.tree')
    else:
        rey_file_name=fprefix+'_reynolds.txt'
        fij_file_name=fprefix+'_fij.txt'
        tree_file_name=fprefix+'_tree.txt'
        treedump_file_name=fprefix+'.tree'
    ## write reynolds to file
    reynolds_out=open(rey_file_name,'w')
    for i in range(D.shape[0]):
        tw=[popnames[i]]
        for j in range(D.shape[1]):
            tw.append(str(D[i,j]))
        print(' '.join(tw), file = reynolds_out)
    reynolds_out.close()
    ## Build the tree
    dist_mat=2*D
    my_nj=nj.NJ(dist_mat,popnames)
    my_nj.fit()
    if outgroup:
        root_edge=[e for e in my_nj.edges if e.n1.label==outgroup or e.n2.label==outgroup][0]
        tree=nj.Rooted_Tree()
        if keep_outgroup:
            tree.build_from_edges(my_nj.edges,root_edge)
            if hzy is not None:
                ## we can optimize root placement
                hzy_dict={}
                for i,nom in enumerate(popnames):
                    hzy_dict[nom]=hzy[i]
                tree.optim_root(hzy_dict)
        else:
            print( "Rooting with %s and dropping it."%outgroup)
            tree.build_from_edges(my_nj.edges,root_edge,outgroup=outgroup)
    else:
        ## No outgroup information
        if hzy is not None:
            hzy_dict={}
            for i,nom in enumerate(popnames):
                hzy_dict[nom]=hzy[i]
            ## we can optimize using heterozygosities
            resid=np.inf ## optim criterion
            root_edge_best=None ## where we will place the root
            for candidate in my_nj.edges:
                edges=[e for e in my_nj.edges]
                tree=nj.Rooted_Tree()
                tree.build_from_edges(edges,candidate)
                tree_fit=tree.optim_root(hzy_dict)
                ##print 'Edge',candidate,' : ',tree_fit.fun
                if tree_fit.x[1]>0 and tree_fit.fun<resid:
                    resid=tree_fit.fun
                    root_edge_best=candidate
                    root_pos_best=tree_fit.x[1]
            tree=nj.Rooted_Tree()
            tree.build_from_edges(my_nj.edges,root_edge_best)
            tree.shift_root(root_pos_best)
        else:
            ## no hzy information, should revert to midpoint but this is
            ## usually not a good idea ...
            raise NotImplementedError
    ## Now we have a tree, get Kinship
    with open(tree_file_name,'w') as f:
        print( tree.newick(), file = f)
    if dump_tree:
        with open(treedump_file_name,'w') as f:
            pickle.dump(tree,f)
    kin,poplabels=tree.kinship()
    with open(fij_file_name,'w') as f:
        for i,nom in enumerate(poplabels):
            print( nom,' '.join([str(x) for x in kin[i,]]), file = f)
    ### we have to be careful because the pop order can change within Fij
    if outgroup and not keep_outgroup:
        pname_temp=[x for x in popnames if x!=outgroup]
    else:
        pname_temp=popnames[:]
    ordre=[poplabels.index(x) for x in pname_temp]
    FF=kin[ordre,:][:,ordre]
    ## clean up temp files if needed
    if not fprefix:
        os.remove(rey_file_name)
        os.remove(fij_file_name)
        os.remove(tree_file_name)
        if dump_tree:
            os.remove(treedump_file_name)
    return FF

class FLK_result():
    def __init__(self):
        self.p0=np.nan
        self.val=0
        self.pval=1
        
class FLK_test():
    '''
    FLK test implementation
    '''
    def __init__(self,kinship,diallelic=True):
        '''
        Creates an FLK test

        Parameters:
        ---------------

        kinship : population kinship matrix
        diallelic : if True, the locus is diallelic (e.g. a SNP) o.w. multiallelic
        
        '''
        self.F=kinship
        assert self.F.shape[0]==self.F.shape[1]
        self.dimension=self.F.shape[0]
        self.diallelic=diallelic
        self.invF=np.linalg.inv(self.F)
        self.un = np.ones(self.dimension)
        self.w = np.dot(self.invF,self.un.T)/np.dot(self.un,np.dot(self.invF,self.un.T))
        self.D,self.Q = np.linalg.eigh(self.F)
        if diallelic:
            self.eigen_contrib=self.eigen_contrib_diallelic
        else:
            self.eigen_contrib=self.eigen_contrib_multi
        
    def eigen_contrib_diallelic(self,p,diallelic=True):
        '''
        Computes orthogonalized contributions of each component to the FLK
        statistic, using the spectral decomposition of the variance
        covariance matrix V(p) V=cste*F=cste*Q*D*Q.T where Q is the matrix
        of eigen vectors and D the corresponding diagonal matrix with
        eigen values on the diagonal.

        Parameters:
        ---------------

        p : vector of population allele frequencies at a locus

        Return Value:
        -----------------
        a tuple of :
            - p0hat : estimated allele frequency in the ancestral population
            - Z : contributions of each PC to the FLK statistic
            - cste : multiplying constant to the test
        the value of FLK is = C*sum(L^2)
        '''
        p0hat = np.dot(self.w.T,p)
        cste = ((1 - (1/(np.dot(self.un,np.dot(self.invF,self.un.T)))))/(p0hat*(1-p0hat)))
        Z=np.dot((p-(p0hat*self.un)),self.Q)
        Z=np.dot(Z,np.diag(np.sqrt(1/self.D)))
        Z *= np.sqrt(cste)
        # tmp=np.dot(p-(p0hat*self.un),self.invF*cste)
        # tmp=np.dot(tmp,(p-(p0hat*self.un).T))
        # print tmp,np.sum([z**2 for z in Z])
        # R = np.dot(self.Q.T,(p-(p0hat*self.un).T))
        # R = np.dot(np.diag(np.sqrt(1/self.D)),R)
        # R *= np.sqrt(cste)
        # print R.shape
        # Z=[R[i] for i in range(R.shape[0])]
        return p0hat,Z,cste

    def eigen_contrib_multi(self,p):
        '''
        Multiallelic version of eigen_contrib_diallelic
        
        Parameters:
        ---------------
        
        p : vector of population allele frequencies at a locus
        
        Return Value:
        -----------------
        a tuple of :
        - p0hat : estimated allele frequency in the ancestral population
        - Z : contributions of each PC to the FLK statistic
        - cste : multiplying constant to the test
        the value of FLK is = C*sum(L^2)
        
        See Also:
        ------------
        
        eigen_contrib_diallelic 
        '''
        p0hat = np.dot(self.w.T,p)
        Z=np.dot((p-(p0hat*self.un)),self.Q)
        Z=np.dot(Z,np.diag(np.sqrt(1/self.D)))
        return p0hat,Z,1.0
    
    def eval_flk(self,p):
        '''
        Computes the FLK statistic

        Parameters:
        ---------------

        p : vector of population allele frequencies

        Return Value:
        -----------------
        a list of :
           -- p0
           -- FLK value
           -- corresponding p-value
           -- each contribution of the PC in turn
        '''
        assert p.shape[0]==self.dimension
        p0,Z,C=self.eigen_contrib(p)
        val=sum([x**2 for x in Z])
        if self.diallelic and p0>0.05 and p0<0.95:
            pval=chi2.sf(val,self.dimension-1)
        else:
            pval=np.nan
        return [p0,val,pval]+Z.tolist()
      



from __future__ import print_function, division
import sys
import numpy as np
import scipy.stats as ss
from scipy.stats import chi2,norm, multivariate_normal
from scipy.stats import percentileofscore
from scipy.optimize import minimize as optim
from scipy import interpolate
from numpy.linalg import multi_dot
from multiprocessing import Pool

#### Base functions for parallel computations (to put into own file)
def pscore(args):
    a, val = args
    return percentileofscore(a,val,kind='weak')

def loglik_h0_qfunc(pzero,p,Vinv):
    ll = - p.shape[0] * np.log( pzero * (1-pzero))
    res = p - pzero
    ll -= multi_dot( (res.T, Vinv, res) ) / (pzero * (1-pzero))
    return -ll

def root_pzero_h0(args):
    p, pini, Vinv = args
    res = optim(loglik_h0_qfunc, pini, args=( p, Vinv), method='SLSQP', bounds = [ (0,1) ],options = {'maxiter':20})
    if not res.success:
        ## we do it by hand
        xx = np.linspace(np.min(p),np.max(p),100) ## pzero should be between min(p) and max(p)
        ll = [ loglik_h0_qfunc( pz, p, Vinv) for pz in xx]
        ill = np.argmin(ll)
        if ill == 0:
            xx = np.linspace(0,np.min(p),100)
            ll =  [ loglik_h0_qfunc( pz, p, Vinv) for pz in xx]
            ill = np.argmin(ll)
        elif ill == 99:
            xx = np.linspace(np.max(p),1,100)
            ll =  [ loglik_h0_qfunc( pz, p, Vinv) for pz in xx]
            ill = np.argmin(ll)
        res = xx[ ill ]
        return (res, 0)
    else:
        return (res.x[0], 1)
                    
def loglik_h1_qfunc(pzero,p,xVinv,jacobian,x,Vinv):
    '''Optimization function to minimize -logLikelihood (i.e. maximize logLikelihood) '''
    #num = np.dot( xVinv, p-pzero)
    beta = multi_dot( (jacobian, xVinv, p-pzero))
    ll = - p.shape[0] * np.log( pzero * (1-pzero))
    res = ( p - pzero - np.dot(x, beta) )
    ll -= multi_dot( (res.T, Vinv, res) ) / (pzero * (1-pzero))
    return -ll

def root_pzero(args):
    p, pini, xVinv, denum, x, Vinv = args
    res = optim(loglik_h1_qfunc,pini, args=(p,xVinv,denum,x,Vinv),method='SLSQP', bounds = [ (0,1) ],options = {'maxiter':20})
    if not res.success:
        ## we do it by hand
        xx = np.linspace(np.min(p),np.max(p),100) ## pzero should be between min(p) and max(p)
        ll = [ loglik_h1_qfunc(pz, p, xVinv,denum,x,Vinv) for pz in xx]
        ill = np.argmin(ll)
        if ill == 0:
            xx = np.linspace(1e-4,np.min(p),100)
            ll =  [ loglik_h1_qfunc( pz, p,xVinv,denum,x, Vinv) for pz in xx]
            ill = np.argmin(ll)
        elif ill == 99:
            xx = np.linspace(np.max(p),0.9999,100)
            ll =  [ loglik_h1_qfunc( pz, p,xVinv,denum,x, Vinv) for pz in xx]
            ill = np.argmin(ll)
        res = xx[ ill ]
        return (res, 0)
    else:
        return (res.x[0], 1)
    
def beta_h1(pzero,p,xVinv, denum):
    num = np.dot( xVinv, (p-pzero) )
    beta = np.dot( denum, num)
    return beta

def calc_beta_h1(args):
    pzero, p, xVinv, denum = args
    return beta_h1(pzero,p,xVinv, denum)

def loglik_h1( pzero, beta, frq, x, Vinv):
    npop = x.shape[0]
    ll   = - npop * np.log( pzero * (1-pzero))
    res  = ( frq - pzero - np.dot(x, beta) )
    ll  -= multi_dot( ( res.T, Vinv, res)) / (pzero * (1-pzero))
    return ll

def calc_loglik_h1(args):
    pzero, beta, p, x, Vinv = args
    return loglik_h1(pzero,beta,p,x,Vinv)

def loglik_h0(p, w, Vinv):
    npop=p.shape[0]
    pzero_hat = np.dot( w.T, p)
    ##pzero_hat = self.xx[ np.argmin( np.abs( pzero_hat - self.xx) ) ]
    ll = - npop * np.log( pzero_hat * (1 - pzero_hat))
    res = (p - pzero_hat)
    ll -= np.dot( res.T, np.dot( Vinv, res ) ) / (pzero_hat * (1 - pzero_hat) )
    return ll

def calc_loglik_h0(args):
    p, w, Vinv = args
    return loglik_h0(p, w, Vinv)

def calc_lrt(args):
    p, xVinv, denum, x, Vinv, w = args
    #### H1 Calculations
    # pini = np.mean(p)
    # res = optim(loglik_h1_qfunc, pini, args=(p, xVinv, denum, x, Vinv), method='SLSQP', bounds = [ (0,1) ],options = {'maxiter':20})
    # if not res.success:
    #     ## we do it by hand
    #     xx = np.linspace(np.min(p),np.max(p),100) ## pzero has to be between min(p) and max(p)
    #     ll = [ loglik_h1_qfunc(pz, p, xVinv,denum,x,Vinv) for pz in xx]
    #     pzero = xx[ np.argmin(ll) ]
    # else:
    #     pzero = res.x[0]
    ## beta
    pbar = np.mean(p)
    if (pbar < 1e-3) or (pbar > 0.999):
        pzero = pbar
        succ1 = 1
        beta = np.zeros(x.shape[1])
        ll1 = 0
        pzero_null=pbar
        ll0 = 0
        succ0 = 1
    else:
        pzero, succ1 = root_pzero( [ p, pbar, xVinv, denum, x, Vinv ])
        beta = multi_dot( (denum, xVinv, p-pzero))
        ll1 = -loglik_h1_qfunc( pzero, p, xVinv, denum, x, Vinv)
        #### H0 Calculations
        pzero_null, succ0 = root_pzero_h0( (p, pbar, Vinv))
        ll0 = -loglik_h0_qfunc( pzero_null, p, Vinv)
    succ = ( ( succ0 == 1 ) and (succ1 ==1) and 1) or 0
    lrt = ll1 - ll0
    if lrt <0 :
        lrt = 0
    res_h0 = { 'llik': ll0, 'pzero': pzero_null, 'pzsucc': succ0}
    res_h1 = { 'llik': ll1, 'pzero': pzero, 'beta': beta, 'pzsucc': succ1}
    return { 'H0': res_h0, 'H1': res_h1, 'lrt':lrt, 'pvalue': chi2.sf( lrt, x.shape[1]),'pzsucc':succ }

def loglik_h0_ML(p, w, Vinv):
    npop=p.shape[0]
    pzero_hat,_ = root_pzero_h0( (p, np.mean(p), Vinv ))
    ll = - npop * np.log( pzero_hat * (1 - pzero_hat))
    res = (p - pzero_hat)
    ll -= multi_dot( ( res.T, Vinv, res)) / (pzero_hat * (1 - pzero_hat) )
    return ll

def calc_lrt_ML_old(args):
    p, xVinv, denum, x, Vinv, w = args
    ## root_pzero
    pini = np.mean(p)
    res = optim(loglik_h1_qfunc, pini, args=(p, xVinv, denum, x, Vinv), method='SLSQP', bounds = [ (0,1) ],options = {'maxiter':20})
    if not res.success:
        ## we do it by hand
        xx = np.linspace(np.min(p),np.max(p),100) ## pzero has to be between min(p) and max(p)
        ll = [ loglik_h1_qfunc(pz, p, xVinv,denum,x,Vinv) for pz in xx]
        pzero = xx[ np.argmin(ll) ]
    else:
        pzero = res.x[0]
    ## beta
    beta = multi_dot( (denum, xVinv, p-pzero))
    ## LRT 
    ll1 = loglik_h1( pzero, beta, p, x, Vinv)
    ll0 = loglik_h0_ML( p, w, Vinv)
    return { 'pzero': np.dot(w.T, p), 'lrt':ll1 - ll0}


def bayes_factor_vect(p, xx, logdet_omega, logdet_nu, logdet_V, beta_mean_fact, omega_inv, Vinv):
    npop = p.shape[0]
    if len(p.shape) == 1:
        frq = p.reshape( npop, 1)
    else: 
        frq = p
    ns = frq.shape[1]
    pzero = np.broadcast_to( xx, ( ns, len(xx)) )

    ## scaling factor
    lp = - npop * np.log( pzero * (1-pzero))
    lp_1 = lp + logdet_omega - logdet_nu - logdet_V
    lp_0 = lp - logdet_V
    ## SNP Sum of Squares
    res = (frq - pzero).T
    res=res.reshape( len(xx), npop, ns)
    postmean = np.einsum( 'ij,kjl->kil', beta_mean_fact, res)
    betass = np.einsum( 'ij,kjl->kil', omega_inv, postmean)
    BT = np.einsum('ijk->ikj',postmean)
    betass = np.einsum( 'ijk,ikj->i',BT,betass)
    datass=np.einsum('ij,kjl->kil',Vinv,res)
    RT=np.einsum('ijk->ikj',res)
    datass=np.einsum('ijk,ikj->i',RT,datass)
    ## compute Bayes Factor
    lp_0 -= datass / (pzero * (1-pzero))
    lp_1 -= ( datass - betass) / ( pzero * ( 1 - pzero))
    vmax0 = np.max(lp_0)
    P0 = np.log( np.sum( np.exp( lp_0 - vmax0) ) ) + vmax0
    vmax1 = np.max(lp_1)
    P1 = np.log( np.sum( np.exp( lp_1 - vmax1) ) ) + vmax1
    return (P1-P0)/np.log(10)
    
def calc_bf(args):
    p, xx, logdet_omega, logdet_nu, logdet_V, beta_mean_fact, omega_inv, Vinv = args
    pbar = np.mean(p)
    if (pbar < 1e-2) or (pbar > 0.99):
        return 0
    return bayes_factor_vect(p, xx, logdet_omega, logdet_nu, logdet_V, beta_mean_fact, omega_inv, Vinv)

def average_log10bf(logbf):
    '''Computes the average of a list of log10(bf)'''
    vmax = np.max(logbf)
    bfm = np.log10( np.mean( np.power(10, logbf - vmax))) + vmax 
    return bfm

def stdize_covar(x):
    try:
        ncovar = x.shape[1]
    except IndexError:
        x = x.reshape( x.shape[0], 1)
        ncovar = 1
    assert len(x.shape) == 2
    sx = np.zeros(x.shape)
    for ic in range(ncovar):
        m = np.mean( x[ :, ic])
        s = np.std( x[ :, ic], ddof=1)
        assert s>0
        sx[:, ic] = (x[ :, ic] - m)/s
    return sx
        
###### MAIN Class
class FLKadapt(object):
    def __init__( self, kinship, x, prefix = 'hapflkadapt', pz_step=0.01, ncpu=4, stdize=True):
        self.prefix=prefix
        self.standardized=False
        if stdize:
            self.standardized = True
            self.x = ss.zscore(x, axis=0, ddof=1) #stdize_covar(x)
        else:
            self.x = x - np.mean(x, axis=0)
            ##print("Covariables not standardized, proceed with care ... ")
        self.F = kinship
        self.diagFbar = np.mean( np.diag( self.F)) ## average distance to root
        _,self.logdet_V = np.linalg.slogdet( self.F)
        self.U,self.S,_ = np.linalg.svd(self.F)
        self.df = self.x.shape[1]
        self.Vinv = np.linalg.inv(self.F) 
        self.xVinv = np.dot(self.x.T,self.Vinv) # x'Vinv
        self.xVinvx = multi_dot( ( self.x.T, self.Vinv, self.x) )
        self.denum = np.linalg.inv( self.xVinvx ) # ( x' Vinv x )^{-1}
        self.npop=self.F.shape[0]
        self.un = np.ones(self.npop)
        self.w = np.dot(self.Vinv,self.un.T)/multi_dot( ( self.un, self.Vinv, self.un.T))
        self.pz_step=pz_step
        self.xx = np.arange(pz_step,1,step=pz_step)
        self.nproc=ncpu
        self.logfile=sys.stdout
        self.pool=None
        ## NULL calibrations
        self.bf_null = None
        self.correct_lrt = None
        
    def __enter__(self):
        self.pool = Pool(self.nproc)
        self.logfile = open(self.prefix+'.log','w')
        return self

    def __exit__(self,*args):
        self.pool.terminate()
        self.logfile.close()
        
    ## LRT Single SNP
    def calibrate_null( self, nsim = 10000, pmin=0.02):
        ''' 
        Create predictor of LRT correcting for inflation on low pzero values.

        1. Simulate frequency data (p) from MVN (pzero, self.F * (p0 * (1-p0))) for 
           p0 in [pmin, 1-pmin], on a log10 scale
        2. Compute LRT(p,x|p0)
        3. Adjust a spline on (x, LRT(p,x|p0)) and create a predictor f(LRT(p,x)) 
           that estimates #df of Chisq | p0
        '''
        assert self.pool is not None
        xp = 10**(np.linspace(np.log10(pmin),np.log10(0.5),20,endpoint=False))
        lrt_res = []
        bf_res = []
        args=[]
        for pzero in xp:
            pm = pzero * np.ones( self.npop)
            mvnorm = multivariate_normal( mean=pm, cov=self.F*pzero*(1-pzero))
            frq = mvnorm.rvs(nsim)
            frq[ frq < 0] = 0
            frq[ frq > 1] = 1
            for isim in range(nsim):
                args.append((frq[isim,], self.xVinv, self.denum, self.x, self.Vinv, self.w))
            # lrt_res += list( np.apply_along_axis( self.fit_loglik_optim, 1, frq, skip_small = False, correct = False))
            bf_res += list(  np.apply_along_axis( self.bayes_factor, 1, frq))
        lrt_res += self.pool.map(calc_lrt, args)
        x = []
        y = []
        for v in lrt_res:
            if not np.isnan(v['lrt']):
              x.append( v['H1']['pzero'])
              x.append( 1-v['H1']['pzero'])
              y.append( v['lrt'])
              y.append( v['lrt'])
        x = np.array(x)
        y = np.array(y)
        idx = np.argsort(x)
        knots = sorted( np.concatenate( [ xp, 1-xp]))[1:-1]
        self.correct_lrt = interpolate.LSQUnivariateSpline(x[idx], y[idx],knots)
        self.bf_null = sorted(bf_res)
        with open(self.prefix+'.calib','w') as f:
            print('# LRT DF',file=f)
            for xx in knots:
                print(xx,self.correct_lrt(xx),file=f)
            print('# BF percentiles',file=f)
            ntot=len(xp)*nsim
            probs=np.array([1.0/ntot, 1.0/nsim, 0.005, 0.01, 0.025, 0.05, 0.1, 0.5, 0.9, 0.95, 0.975, 0.99, 0.995, 1-1.0/nsim, 1-1.0/ntot])
            pc = np.percentile(bf_res, q=100*probs)
            for i,q in enumerate(probs):
                print(q,pc[i],file=f)
        
    def beta_h1(self,pzero,p):
        num = np.dot( self.xVinv, (p-pzero) )
        beta = np.dot(self.denum, num)
        return beta
    
    def loglik_h1(self,pzero,p):
        beta = self.beta_h1(pzero,p)
        ll = - self.npop * np.log( pzero * (1-pzero))
        res = ( p - pzero - np.dot(self.x, beta) )
        ll -= np.dot( res.T, np.dot( self.Vinv, res) ) / (pzero * (1-pzero))
        return {'llik': ll, 'beta': beta}
    
    def loglik_h0(self,p):
        pzero_hat = np.dot( self.w.T, p)
        ##pzero_hat = self.xx[ np.argmin( np.abs( pzero_hat - self.xx) ) ]
        ll = - self.npop * np.log( pzero_hat * (1 - pzero_hat))
        res = (p - pzero_hat)
        ll -= np.dot( res.T, np.dot( self.Vinv, res ) ) / (pzero_hat * (1 - pzero_hat) )
        return {'llik': ll, 'pzero': pzero_hat}

    def fit_loglik(self,p):
        res_h0 = self.loglik_h0(p)
        profile_h1 = [ self.loglik_h1(pz,p) for pz in self.xx ]
        best_pz = np.argmax( [ x['llik'] for x in profile_h1 ] )
        res_h1 = profile_h1[ best_pz ]
        pzero_hat = self.xx[best_pz]
        ## var_beta = self.denum*pzero_hat*(1-pzero_hat) # does'nt seem to work here 
        res_h1.update( { 'pzero' : self.xx[best_pz] } )
        lbd= (res_h1['llik']-res_h0['llik'])
        if lbd<0:
            lbd=np.nan
        return {'H0': res_h0, 'H1': res_h1 ,'Lbd': lbd, 'pval': chi2.sf(lbd,self.df)}

    def fit_loglik_optim_ML(self, p, skip_small=True):
        res_h0 = self.loglik_h0(p)
        pz=res_h0['pzero']
        if skip_small and (np.min([pz,1-pz]) < 0.05):
            res_h1={ 'pzero': np.nan, 'llik': np.nan, 'beta': np.full((self.df),np.nan)}
            lbd=np.nan
        else:
            val_h0,_ = root_pzero_h0( (p, res_h0['pzero'], self.Vinv))
            res_h0['pzero'] = val_h0
            res_h0['llik'] = -loglik_h0_qfunc( val_h0, p, self.Vinv)
            val_h1,_ = root_pzero( ( p, val_h0, self.xVinv,self.denum,self.x,self.Vinv))
            res_h1 = self.loglik_h1(val_h1,p)
            res_h1.update({ 'pzero' : val_h1} )
            lbd= (res_h1['llik']-res_h0['llik'])
            ##print(pz, res_h0['pzero'], res_h1['pzero'], res_h0['llik'], res_h1['llik'], file=self.logfile)
            # if correct:
            #     df = self.correct_lrt( res_h0['pzero'])
            #     lbd = chi2.ppf( chi2.cdf( lbd, df = df), df = self.df)
            ##print(res_h0['pzero'],val_h1.x[0],val_h1.success,val_h1.nfev,val_h1.nit,lbd)
        return {'H0': res_h0, 'H1': res_h1 ,'lrt': lbd, 'pvalue': chi2.sf(lbd,self.df)}

    def FLKadapt(self, frq, skip_small = True):
        try:
            npop, nsnp = frq.shape
        except:
            return self.fit_loglik_optim_ML(p, frq, skip_small)
        args=[]
        for isnp in range(nsnp):
                args.append((frq[:,isnp], self.xVinv, self.denum, self.x, self.Vinv, self.w))
        lrt_res = self.pool.map(calc_lrt, args)
        return lrt_res
  
    def fit_loglik_optim(self,p,skip_small=True, correct = True):
        ''' Numerical optimization of loglikelihood'''
        res_h0 = self.loglik_h0(p)
        pz=res_h0['pzero']
        if skip_small and (np.min([pz,1-pz]) < 0.05):
            res_h1={ 'pzero': np.nan, 'llik': np.nan, 'beta': np.full((self.df),np.nan)}
            lbd=np.nan
        else:
            val_h1 = optim(loglik_h1_qfunc,res_h0['pzero'], args=(p,self.xVinv,self.denum,self.x,self.Vinv),method='SLSQP', bounds = [ (0,1) ],options = {'maxiter':20})
            res_h1 = self.loglik_h1(val_h1.x[0],p)
            res_h1.update({ 'pzero' : val_h1.x[0]} )
            lbd= (res_h1['llik']-res_h0['llik'])
            if correct:
                df = self.correct_lrt( res_h0['pzero'])
                lbd = chi2.ppf( chi2.cdf( lbd, df = df), df = self.df)
            ##print(res_h0['pzero'],val_h1.x[0],val_h1.success,val_h1.nfev,val_h1.nit,lbd)
        return {'H0': res_h0, 'H1': res_h1 ,'Lbd': lbd, 'pvalue': chi2.sf(lbd,self.df)}

    #### hapFLKadapt
    def PCA_clust(self,kfrq):
        E,K,R = kfrq.shape
        Pbar = np.einsum('j,klj->kl',self.w,kfrq)
        Pbar=np.kron(Pbar,np.ones(R)).reshape(E,K,R)
        Pmat = np.concatenate(kfrq-Pbar).T
        assert Pmat.shape == (R,E*K)
        u,s,v = np.linalg.svd(Pmat)
        print('Global Lbd',*self.S)
        print('Local Lbd', *s)
        with open('test_Qglob.txt','w') as f:
            for r in range(R):
                print(*self.U[r,],file=f)
        with open('test_Qloc.txt','w') as f:
            for r in range(R):
                print(*u[r,],file=f)
        for pc in range(R):
            for r in range(R):
                print(pc,u[r,pc],self.U[r,pc],self.x[r,0],file=self.logfile)
    
    def BF_clust(self,kfrq, pmin = 0.05, beta_priors = None):
        if beta_priors is None:
            beta_priors = np.power( 10, np.linspace( -1, 0, 10))
        
        assert self.pool is not None
        E, K, R = kfrq.shape
        P = len( beta_priors)
        args = []
        Pbar = np.mean( kfrq, axis = 2)
        subset = np.array( ( Pbar > pmin) & ( Pbar < (1-pmin)))
        Keff = np.sum( subset, axis = 1)
        ## compute logBF_k for each EM 
        for e in range(E):
            for k in range(K):
                ## ajout boucle prior
                for ip, nu in enumerate( beta_priors):
                    self.set_beta_prior( var_prior = nu * self.diagFbar)
                    args.append( ( kfrq[e,k,], self.xx, self.logdet_omega, self.logdet_nu, self.logdet_V, self.beta_mean_fact, self.omega_inv, self.Vinv))
        logBFk =  np.array( self.pool.map( calc_bf, args))
        logBFk = logBFk.reshape( E, K, P)
        # print("New locus")
        # with open('bf.profile','w') as fbf:
        #     print('e','k','pz','nu','Keff','logbf',file=fbf)
        #     for e in range(E):
        #         for k in range(K):
        #             if pmin < Pbar[e,k] < (1 - pmin) :
        #                 for ip, nu in enumerate( beta_priors):
        #                     print( e, k, Pbar[ e, k], nu, Keff[e], logBFk[ e, k, ip],file=fbf)
                    
        ## only consider clusters which pmin < mean(pk) < pmin
        logBFe = np.zeros(E)
        for e in range(E):
            subK = range(K)
            for k in subK:
                if pmin < Pbar[ e, k] < (1-pmin):
                    logbfk_i =  average_log10bf( logBFk[ e, k, ])
                    logBFe[e] += logbfk_i #average_log10bf( logBFk[ e, k, ] )
            ##print( e, logBFe[ e])
            # if Keff > 0:
            #     ## uniform prior on number of associated clusters
            #     logBFe[e] += Keff * np.log10( (Keff + 1)/( 2*Keff))
        return average_log10bf( logBFe)

    
    def hapflkadapt(self, kfrq):
        assert self.pool is not None
        assert self.correct_lrt is not None
     
        if len(kfrq.shape)==4:
            E, K, R, L = kfrq.shape
        elif len(kfrq.shape) == 3:
            return self.fit_loglik_clust_pz(kfrq)
        else:
            raise ValueError("Bad Shape for kfrq")
        
        LRT  = np.zeros( ( E, K, L))
        Succ = np.zeros( ( E, K, L))
        args = []
        for e in range(E):
            for k in range(K):
                for l in range(L):
                    args.append( ( kfrq[e,k,:,l], self.xVinv, self.denum, self.x, self.Vinv, self.w))
        res =  self.pool.map( calc_lrt, args)
        
        LRT = np.array( [ x['lrt'] for x in res])
        Succ = np.array([ x['pzsucc'] for x in res])
        
        LRT = LRT.reshape(E,K,L)
        Succ = Succ.reshape(E,K,L)
        LRT = LRT*Succ

        hflka_res = []
        for l in range(L):
            myLRT = LRT[:,:,l]
            mykfrq = kfrq[:,:,:,l]
            Pbar = np.einsum('j,klj->kl',self.w,mykfrq)
            subset = np.array( ( Pbar > 0.05) & ( Pbar < 0.95))
            df_pbar = self.correct_lrt(Pbar)
            df = np.sum( df_pbar*subset , axis=1)
            lrt = np.sum( myLRT*subset, axis=1)
            qchisq = chi2.sf(lrt, df)
            zscores = -norm.ppf( qchisq)
            Ztot = np.mean(zscores)
            hflka_res.append({ 'Z':Ztot, 'pvalue': norm.sf(Ztot), 'Z_mat': zscores})
        return hflka_res

    def fit_loglik_clust_pz(self,kfrq):
        '''
        Computes Likelihood Ratio Test of model:

        kfrq = p0_k + X beta_k + e 
        e ~ N( 0, F. p0_k_1*(1-p0_k_1))

        against:

        kfrq = p0_k + e 
        e ~ N( 0, F. p0_k_0*(1-p0_k_0))
        input:
            -- kfrq: cluster frequencies ( E x K x R ) 
            with E (Number of EM fits), K (number of clusters), R (number of pops)
        '''
        assert self.pool is not None
        assert self.correct_lrt is not None
        E, K, R = kfrq.shape
        args = []
        for e in range(E):
            for k in range(K):
                args.append( ( kfrq[e,k,], self.xVinv, self.denum, self.x, self.Vinv, self.w))
        res =  self.pool.map( calc_lrt, args)
        LRT = np.array( [ x['lrt'] for x in res])
        Succ = np.array([ x['pzsucc'] for x in res])

        LRT = LRT.reshape(E,K)
        Succ = Succ.reshape(E,K)
            
        LRT = LRT*Succ
        
        Pbar = np.einsum('j,klj->kl',self.w,kfrq)
        subset = np.array( ( Pbar > 0.05) & ( Pbar < 0.95))
        df_pbar = self.correct_lrt(Pbar)
        df = np.sum( df_pbar*subset , axis=1)
        # df = np.sum( subset , axis=1)
        lrt = np.sum( LRT*subset, axis=1)
        qchisq = chi2.sf(lrt, df)
        zscores = -norm.ppf( qchisq)
        Ztot = np.mean(zscores)
        return({ 'Z':Ztot, 'pvalue': norm.sf(Ztot), 'Z_mat': zscores})
  
    def fit_loglik_clust(self,kfrq):
        '''
        Computes Likelihood Ratio Test of model :

        kfrq = p0_k + X beta_k + e 
        e ~ N( 0, F . var_k_1)
        
        against:
        
        kfrq = p0_k + e
        e ~ N(0, F . var_k_0)

        input:
        -- kfrq: cluster frequencies ( E x K x R ) 
        with E (Number of EM fits), K (number of clusters), R (number of pops)
        '''
        E,K,R = kfrq.shape
        P = self.df
        pk = kfrq.reshape( ( E, K, R, 1))
        ## MLE  under H1
        unx = np.hstack( (self.un.reshape(R,1), self.x))
        ### ( X' . Vinv . X)inv . (X' . Vinv) p 
        Emat =  np.linalg.inv( multi_dot( [unx.T, self.Vinv, unx]) )
        Emat = multi_dot( [ Emat, unx.T, self.Vinv] )
        mles = np.einsum('ij,kljm->klim',Emat,pk)
        ### yhat = X . B
        fitted = np.einsum('ij,kljm->klim',unx,mles)
        resid = pk - fitted
        ### (y - yhat)' Vinv (y - yhat)
        RSS = np.einsum('ij,kljm->klim',self.Vinv,resid)
        RSS = np.einsum('ijkl,ijkm->ijlm',resid,RSS)
        mle_sigma=RSS/R
        ll_h1 = -R*np.log(mle_sigma) - RSS/mle_sigma
        ## MLE under H0
        _un = self.un.reshape(R,1)
        Emat = np.linalg.inv( multi_dot( [_un.T, self.Vinv, _un]) )
        Emat = multi_dot( [ Emat, _un.T, self.Vinv] )
        mles_H0 = np.einsum('ij,kljm->klim',Emat,pk)
        fitted_H0 = np.einsum('ij,kljm->klim',_un,mles_H0)
        resid_H0 = pk - fitted_H0
        RSS_H0 = np.einsum('ij,kljm->klim',self.Vinv,resid_H0)
        RSS_H0 = np.einsum('ijkl,ijkm->ijlm',resid_H0,RSS_H0)
        mle_sigma_H0 = RSS_H0/R
        ll_h0 = -R*np.log(mle_sigma_H0) - RSS_H0/mle_sigma_H0

        LRT = (ll_h1 - ll_h0).reshape(E,K)
        LRT_av = np.mean(np.sum(LRT,axis=1))
        return({ 'Lbd':LRT_av, 'pvalue': chi2.sf(LRT_av,df=K), 'LRT_mat': LRT})
        
    ## Bayes Factors
    def set_beta_prior(self,var_prior=0.1):
        self.prior = True
        self.nu = var_prior
        ## 1 par
        self.nuinv = np.diag([1/self.nu]*self.df)
        self.logdet_nu = self.df * np.log(self.nu) 
        self.omega_inv = self.nuinv + self.xVinvx
        self.omega = np.linalg.inv(self.omega_inv)
        self.beta_mean_fact = np.dot( self.omega, self.xVinv)
        _,self.logdet_omega = np.linalg.slogdet( self.omega)
   
    def bayes_factor_vect(self,p):
        if len(p.shape) == 1:
            frq = p.reshape( self.npop, 1)
        else: 
            frq = p
        ns = frq.shape[1]
        pzero = np.broadcast_to( self.xx, ( ns, len(self.xx)) )
        
        ## scaling factor
        lp = - self.npop * np.log( pzero * (1-pzero))
        lp_1 = lp + self.logdet_omega - self.logdet_nu - self.logdet_V
        lp_0 = lp - self.logdet_V
        ## SNP Sum of Squares
        res = (frq - pzero).T
        res=res.reshape( len(self.xx), self.npop, ns)
        postmean = np.einsum( 'ij,kjl->kil', self.beta_mean_fact, res)
        betass = np.einsum( 'ij,kjl->kil', self.omega_inv, postmean)
        BT = np.einsum('ijk->ikj',postmean)
        betass = np.einsum( 'ijk,ikj->i',BT,betass)
        datass=np.einsum('ij,kjl->kil',self.Vinv,res)
        RT=np.einsum('ijk->ikj',res)
        datass=np.einsum('ijk,ikj->i',RT,datass)
        ## compute Bayes Factor
        lp_0 -= datass / (pzero * (1-pzero))
        lp_1 -= ( datass - betass) / ( pzero * ( 1 - pzero))
        vmax0 = np.max(lp_0)
        P0 = np.log( np.sum( np.exp( lp_0 - vmax0) ) ) + vmax0
        vmax1 = np.max(lp_1)
        P1 = np.log( np.sum( np.exp( lp_1 - vmax1) ) ) + vmax1
        return (P1-P0)/np.log(10)

    def bayes_factor(self,p, beta_priors = None ):
        ## priors: variance of divergence in scale of drift units
        ## assumes covariables are standardized.
        if beta_priors is None:
            beta_priors = np.power(10,np.linspace(-1,1,10))
        bf = np.zeros(len(beta_priors))
        for i,nu in enumerate(beta_priors):
            self.set_beta_prior( var_prior = nu * self.diagFbar )
            bf[i] =  self.bayes_factor_vect(p)
        vmax = np.max(bf)
        bf = np.log10( np.sum( np.power(10, bf - vmax))) + vmax
        return bf - np.log10( len(beta_priors))

    def bayesFLKadapt( self, frq, beta_priors = np.power( 10, np.linspace(-1,1,10) )):
        npop,nsnp = frq.shape
        P = len(beta_priors)
        args = []
        for ip, nu in enumerate( beta_priors):
            self.set_beta_prior( var_prior = nu*self.diagFbar)
            for s in range(nsnp):
                args.append( (frq[:, s], self.xx, self.logdet_omega, self.logdet_nu, self.logdet_V, self.beta_mean_fact, self.omega_inv, self.Vinv))
        logBF_t = np.array( self.pool.map( calc_bf, args))
        logBF_t = logBF_t.reshape( P, nsnp)
        logBF = self.pool.map( average_log10bf, logBF_t.T)
        return logBF
            
        
    def bf_rank(self,val):
        assert self.pool is not None
        val = np.asarray(val)
        args=[]
        for x in val:
            args.append((self.bf_null,x))
        scores = self.pool.map(pscore, args)
        scores = np.array( scores )
        p = 1 - scores/100
        p[ p==0] = 0.5/len(self.bf_null)
        p[ p==1] =  1-0.5/len(self.bf_null)
        return p
        
    def logprob(self,pzero,p):
        ''' Computes log(p|pzero,x,Beta) under Beta=0 and Beta>0 
        Returns:
            lp = [logp0,logp1] 
        '''
        ## scaling factor
        lp = - self.npop * np.log( pzero * (1-pzero))
        lp_1 = lp + self.logdet_omega - self.logdet_nu - self.logdet_V
        lp_0 = lp - self.logdet_V

        ## SNP Sum of Squares
        res = (p - pzero)
        postmean = np.dot( self.beta_mean_fact, res )
        betass = multi_dot( ( postmean.T, self.omega_inv, postmean) )
        datass = multi_dot( ( res.T, self.Vinv, res ) )

        lp_0 -= datass / ( pzero * ( 1 - pzero))
        lp_1 -= ( datass - betass) / ( pzero * ( 1 - pzero))
        return np.array([lp_0,lp_1])

 
        
# if __name__ == '__main__':    
#     #myh=HapFLK.from_db_file('../hapflkadapt/covkin.db')
#     ##myh=HapFLK.from_db_file('./hapflk.db')
#     #myh=HapFLK.from_db_file('../hapflkadapt/bta29.db')
#     ##myh.kinship=myh.kinship+0.01*(np.trace(myh.kinship)/myh.npop)

    
#     # input_file='../hapflkadapt/bta29.db'
#     try:
#         input_file=sys.argv[1]
#         output_file_prefix=input_file[:-3]+'_test'
#     except IndexError:
#         ##input_file='../hapflkadapt/BTA2.db'
#         input_file='../hapflkadapt/bta29.db'
#         output_file_prefix='test'
#         print(input_file,output_file_prefix)
#         myh=HapFLK.from_db_file(input_file)
#         nullcov = norm.rvs(size=myh.npop)
#         myh.counter.new('NULL Analysis')
#         Covs=np.vstack([nullcov]).T
#         with FLKadapt(myh.kinship, Covs, prefix='test',ncpu=16) as ffnull:
#             ##ffnull.calibrate_null(nsim=1000)
#             test_snps = range(0,myh.nsnp,1)
#             myh.counter.new('hapFLKadapt')
#             if myh.kfrq is not None:
#                 #pass
#                 # resnull_h = [ffnull.PCA_clust(myh.kfrq[:,:,:,i]) for i in range(200,myh.nsnp,20000)]
#                 resnull_h = [ffnull.fit_loglik_clust_pz(myh.kfrq[:,:,:,i]) for i in test_snps]
#                 # resnull_h = [ffnull.BF_clust(myh.kfrq[:,:,:,i]) for i in range(0,myh.nsnp,200)]
#             # myh.counter.new('FLKadapt: BF Stat 2 pars')
#             # bfnull2 = np.apply_along_axis(ffnull.bayes_factor_2pars,0,myh.frq)
#             # myh.counter.new('LRT raw')
#             # lrtraw = np.apply_along_axis(ffnull.fit_loglik_optim,0,myh.frq, correct = False)
#             # myh.counter.new('LRT cor')
#             # lrtcor = np.apply_along_axis(ffnull.fit_loglik_optim,0,myh.frq)
#             myh.counter.new('LRT ML')
#             lrtml = np.apply_along_axis(ffnull.fit_loglik_optim_ML,0,myh.frq)
#             myh.counter.new('BF')
#             bfnull  = np.apply_along_axis(ffnull.bayes_factor,0,myh.frq)
#             # print(bfnull2.shape,lrtnull.shape,bfnull.shape)
#             with open( 'test.out', 'w') as fout:
#                 print( 'LRTr', 'pvalr', 'LRTc', 'pvalc', 'LRTML','pvalML', 'bf',  'Zh', 'pvalh', file=fout)
#                 for ix,i in enumerate(test_snps):
#                     print( lrtraw[i]['Lbd'], lrtraw[i]['pvalue'], lrtcor[i]['Lbd'], lrtcor[i]['pvalue'], lrtml[i]['Lbd'], lrtml[i]['pvalue'], bfnull[i], resnull_h[ix]['Z'], resnull_h[ix]['pvalue'], file=fout)
#         myh.counter.end()
#         sys.exit(0)
            
#     print(input_file,output_file_prefix)
#     myh=HapFLK.from_db_file(input_file)
    
#     covdata=pd.read_table('../data/breed_phenotypes.txt',sep=' ')
#     covpops=list(covdata.index)
#     smscov=np.array( [ covdata.values[ covpops.index(p), 1] for p in myh.pops])
#     smscov=(smscov-np.mean(smscov))/ np.std(smscov)
#     prodcov=np.array( [ covdata.values[ covpops.index(p), 0] == 'Dairy' for p in myh.pops], dtype=np.int)
    
#     ## Stature
#     myh.counter.new('Stature Analysis')
#     Covs=np.vstack([smscov]).T
#     with FLKadapt(myh.kinship, Covs, prefix = output_file_prefix, ncpu=16) as ffsms:
#         myh.counter.new('Null Simulations')
#         ffsms.calibrate_null()
#         myh.counter.new('FLKadapt: LRT Stat')
#         ressms = np.apply_along_axis(ffsms.fit_loglik_optim_ML,0,myh.frq)
#         myh.counter.new('FLKadapt: BF Stat')
#         bfsms = np.apply_along_axis(ffsms.bayes_factor,0,myh.frq)
#         rankbfsms = ffsms.bf_rank(bfsms)
#         if myh.kfrq is not None:
#             myh.counter.new('hapFLKadapt')
#             ressms_h = [ffsms.fit_loglik_clust_pz(myh.kfrq[:,:,:,i]) for i in range(myh.nsnp)]
         
#     ## Write out
#     myh.counter.new('Writing results')
#     with open(output_file_prefix+'_sms_opt.out','w') as fout:
#         print('rs','chr','pos','pzero.null','bf', 'bfrank', 'LRT', 'pvalue', 'Zh', 'pvalueh', 'pzero','beta', file=fout)
#         for i,d in enumerate(ressms):
#             rs=myh.sorted_snps[i].name
#             chrom,_,pos=myh.carte.position(rs)
#             print( rs, chrom, int(pos), d['H0']['pzero'], bfsms[i], rankbfsms[i],  d['Lbd'],  d['pvalue'], ressms_h[i]['Z'], ressms_h[i]['pvalue'], d['H1']['pzero'], *d['H1']['beta'],  file=fout)

#     ## Production
#     myh.counter.new('Type Analysis')
#     Covs=np.vstack([prodcov]).T
#     with FLKadapt(myh.kinship, Covs, prefix = output_file_prefix, ncpu=16) as ffprod:
#         myh.counter.new('Null Simulations')
#         ffprod.calibrate_null()
#         myh.counter.new('FLKadapt: LRT Stat')
#         resprod = np.apply_along_axis( ffprod.fit_loglik_optim_ML, 0, myh.frq)
#         myh.counter.new('FLKadapt: BF Stat')
#         bfprod = np.apply_along_axis( ffprod.bayes_factor, 0, myh.frq)
#         rankbfprod = ffprod.bf_rank(bfprod)
#         if myh.kfrq is not None:
#             myh.counter.new('hapFLKadapt')
#             resprod_h = [ffprod.fit_loglik_clust_pz(myh.kfrq[:,:,:,i]) for i in range(myh.nsnp)]

  
#     myh.counter.new('Writing results')
#     with open(output_file_prefix+'_prod_opt.out','w') as fout:
#         print('rs','chr','pos','pzero.null','bf', 'bfrank', 'LRT', 'pvalue', 'Zh', 'pvalueh', 'pzero','beta', file=fout)
#         for i,d in enumerate(resprod):
#             rs=myh.sorted_snps[i].name
#             chrom,_,pos=myh.carte.position(rs)
#             print( rs, chrom, int(pos), d['H0']['pzero'], bfprod[i], rankbfprod[i], d['Lbd'],  d['pvalue'], resprod_h[i]['Z'], resprod_h[i]['pvalue'], d['H1']['pzero'], *d['H1']['beta'],  file=fout)
#     print()
 
#     myh.counter.end()
